from typing import Literal, Union

from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import SKLearnModelBase

try:
    from sklearn.svm import SVR
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
This is an sklearn support vector regressor model.

The objective of a support vector machine algorithm is to find a hyperplane in an
n-dimensional space that best separates the data. The data points on
either side of the hyperplane that are closest to the hyperplane are called Support
Vectors. These influence the position and orientation of the hyperplane and thus help
build the SVM.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
kernel : {'linear', 'poly', 'rbf', 'sigmoid', 'precomputed'} or callable,  \
        default='rbf'
    Specifies the kernel type to be used in the algorithm.
    If none is given, 'rbf' will be used. If a callable is given it is
    used to precompute the kernel matrix.
degree : int, default=3
    Degree of the polynomial kernel function ('poly').
    Ignored by all other kernels.
gamma : {'scale', 'auto'} or float, default='scale'
    Kernel coefficient for 'rbf', 'poly' and 'sigmoid'.
    - if ``gamma='scale'`` (default) is passed then it uses
      1 / (n_features * X.var()) as value of gamma,
    - if 'auto', uses 1 / n_features.
coef0 : float, default=0.0
    Independent term in kernel function.
    It is only significant in 'poly' and 'sigmoid'.
tol : float, default=1e-3
    Tolerance for stopping criterion.
C : float, default=1.0
    Regularization parameter. The strength of the regularization is
    inversely proportional to C. Must be strictly positive.
    The penalty is a squared l2 penalty.
epsilon : float, default=0.1
     Epsilon in the epsilon-SVR model. It specifies the epsilon-tube
     within which no penalty is associated in the training loss function
     with points predicted within a distance epsilon from the actual
     value.
shrinking : bool, default=True
    Whether to use the shrinking heuristic.
cache_size : float, default=200
    Specify the size of the kernel cache (in MB).
verbose : bool, default=False
    Enable verbose output. Note that this setting takes advantage of a
    per-process runtime setting in libsvm that, if enabled, may not work
    properly in a multithreaded context.
max_iter : int, default=-1
    Hard limit on iterations within solver, or -1 for no limit.
"""

Kernel = Literal["linear", "poly", "rbf", "sigmoid", "precomputed"]
Gamma = Union[Literal["scale", "auto"], float]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class SupportVectorRegressorConfig(ModelConfig):
    kernel: Kernel = "rbf"
    degree: int = 3
    gamma: Gamma = "scale"
    coef0: float = 0
    tol: float = 1e-3
    C: float = 1
    epsilon: float = 0
    shrinking: bool = True
    cache_size: float = 200
    verbose: bool = False
    max_iter: int = -1


class SupportVectorRegressor(SKLearnModelBase[SupportVectorRegressorConfig]):
    @property
    def _config_builder(self) -> type[SupportVectorRegressorConfig]:
        return SupportVectorRegressorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> SVR:
        config = self.model_config
        return SVR(
            kernel=config.kernel,
            degree=config.degree,
            gamma=config.gamma,
            coef0=config.coef0,
            tol=config.tol,
            C=config.C,
            epsilon=config.epsilon,
            shrinking=config.shrinking,
            cache_size=config.cache_size,
            verbose=config.verbose,
            max_iter=config.max_iter,
        )



"""Mean squared error regression loss."""

import logging
from typing import Any

import evaluate
from sklearn.metrics import mean_squared_error

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
The mean_squared_error function computes mean square error, a risk metric
corresponding to the expected value of the squared (quadratic) error or loss.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated target values.
    references: Ground truth (correct) target values.
    sample_weight (optional): Weighting of each sample.
    multioutput (optional): Defines aggregating of multiple output values.
        Array-like value defines weights used to average errors. Alternatively,
        one of {'raw_values', 'uniform_average'}. Defaults to 'uniform_average'.
        'raw_values' :
            Returns a full set of errors in case of multioutput input.
        'uniform_average' :
            Errors of all outputs are averaged with uniform weight.

Returns:
    mean_squared_error: A non-negative floating point value (the best value
    is 0.0), or an array of floating point values, one for each individual target.

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("mean_squared_error")
    >>> predictions = [2.5, 0.0, 2, 8]
    >>> references = [3, -0.5, 2, 7]
    >>> metric.compute(predictions=predictions, references=references)
    {'mean_squared_error': 0.375}
    >>> references = [2.5, 0.0, 2, 8]
    >>> predictions = [3, -0.5, 2, 7]
    >>> metric = load_metric("mean_squared_error", config_name="multioutput")
    >>> predictions = [[0, 2],[-1, 2],[8, -5]]
    >>> references = [[0.5, 1],[-1, 1],[7, -6]]
    >>> metric.compute(predictions=predictions, references=references)
    {'mean_squared_error': 0.708...}
    >>> metric.compute(predictions=predictions, references=references, multioutput='raw_values')
    {'mean_squared_error': array([0.416..., 1.        ])}
    >>> metric.compute(predictions=predictions, references=references, multioutput=[0.3, 0.7])
    {'mean_squared_error': 0.825...}
"""

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class MeanSquaredError(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("float")),
                    "references": datasets.Sequence(datasets.Value("float")),
                }
                if self.config_name == "multioutput"
                else {
                    "predictions": datasets.Value("float"),
                    "references": datasets.Value("float"),
                },
            ),
            reference_urls=[
                "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        sample_weight: list[float] | None = None,
        multioutput: str = "uniform_average",
        **kwargs: Any,
    ) -> MetricResult:
        score = mean_squared_error(
            y_true=references,
            y_pred=predictions,
            sample_weight=sample_weight,
            multioutput=multioutput,
        )

        return {self.tag: score}



import pytest

from molflux.splits.presets import k_fold_split
from molflux.splits.strategies.core.group_k_fold import GroupKFold
from molflux.splits.strategies.core.k_fold import KFold
from molflux.splits.strategies.core.stratified_k_fold import StratifiedKFold


def test_returns_k_fold_split_by_default():
    """That a ShuffleSplit strategy is returned by default"""
    strategy = k_fold_split()
    assert isinstance(strategy, KFold)


def test_returns_k_fold_if_not_stratified_and_not_grouped():
    """That a KFold strategy is returned if 'stratified' is False, and 'grouped' is False."""
    strategy = k_fold_split(stratified=False, grouped=False)
    assert isinstance(strategy, KFold)


def test_returns_group_k_fold_if_not_stratified_and_grouped():
    """That a GroupKFold strategy is returned if 'stratified' is False, and 'grouped' is True."""
    strategy = k_fold_split(stratified=False, grouped=True)
    assert isinstance(strategy, GroupKFold)


def test_returns_stratified_k_fold_if_stratified_and_not_grouped():
    """That a StratifiedKFold strategy is returned if 'stratified' is True, and 'grouped' is False."""
    strategy = k_fold_split(stratified=True, grouped=False)
    assert isinstance(strategy, StratifiedKFold)


def test_raises_on_not_implemented_combination():
    """That an exception is raised if requesting an unsupported combination of
    'stratified', and 'grouped' parameters."""
    with pytest.raises(NotImplementedError):
        k_fold_split(stratified=True, grouped=True)



from typing import Any

import torch
from torch.utils.data import Dataset

import datasets
from molflux.modelzoo.models.lightning.datamodule import LightningDataModule
from molflux.modelzoo.models.lightning.mlp_regressor.mlp_config import MLPConfig


class MLPDataset(Dataset):
    def __init__(
        self,
        dataset: datasets.Dataset,
        x_features: list[str],
        y_features: list[str],
        with_y_features: bool,
    ) -> None:
        self.dataset = dataset
        self.x_features = x_features
        self.y_features = y_features
        self.with_y_features = with_y_features

        all_features = self.x_features
        if with_y_features:
            all_features = all_features + self.y_features

        self.dataset.set_format("torch", columns=all_features)

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, idx: int) -> tuple[torch.Tensor, torch.Tensor]:
        datapoint = self.dataset[idx]

        x_list: list[torch.Tensor] = []
        for x_col in self.x_features:
            x_feat = datapoint[x_col]

            if len(x_feat.shape) == 0:
                x_feat = x_feat.unsqueeze(0)
            elif len(x_feat.shape) > 1:
                raise RuntimeError(
                    f"The shape of x_feature {x_col} = {x_feat.shape} cannot be concatenated.",
                )

            x_list.append(x_feat)

        x = torch.cat(x_list, dim=0)

        if self.with_y_features:
            y_list: list[torch.Tensor] = []
            for y_col in self.y_features:
                y_feat = datapoint[y_col]

                if len(y_feat.shape) == 0:
                    y_feat = y_feat.unsqueeze(0)
                elif len(y_feat.shape) > 1:
                    raise RuntimeError(
                        f"The shape of y_feature {y_col} = {y_feat.shape} cannot be concatenated.",
                    )

                y_list.append(y_feat)

            y = torch.cat(y_list, dim=0)
        else:
            y = torch.empty(0)

        return x, y


class MLPDataModule(LightningDataModule):
    model_config: MLPConfig

    def prepare_dataset(
        self,
        data: datasets.Dataset,
        split: str,
        name: str | None = None,
        **kwargs: Any,
    ) -> Dataset:
        del name, kwargs

        with_y_features = split != "predict"

        return MLPDataset(
            dataset=data,
            x_features=self.model_config.x_features,
            y_features=self.model_config.y_features,
            with_y_features=with_y_features,
        )



from collections.abc import Iterable, Iterator
from typing import Any

import sklearn.model_selection

from molflux.splits.bases import SplittingStrategyBase
from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable

_DESCRIPTION = """
Leave One Group Out cross-validator

Provides train/validation indices to split data according to a third-party
provided group. This group information can be used to encode arbitrary
domain specific stratifications of the samples as integers.

For instance the groups could be the year of collection of the samples
and thus allow for cross-validation against time-based splits.

Notes
-----
Splits are ordered according to the index of the group left out. The first
split has validation set consisting of the group whose index in `groups` is
lowest, and so on.
"""


class LeaveOneGroupOut(SplittingStrategyBase):
    def _info(self) -> SplittingStrategyInfo:
        return SplittingStrategyInfo(
            description=_DESCRIPTION,
        )

    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Args:
            dataset: The data to be split.
            groups: Group labels for the samples used while splitting the dataset.

        Examples:
            >>> from molflux.splits import load_splitting_strategy
            >>> strategy = load_splitting_strategy('leave_one_group_out')
            >>> dataset = range(10)
            >>> groups = [1, 1, 2, 2, 2, 3, 3, 3, 3]
            >>> folds = strategy.split(dataset, groups=groups)
        """

        # Wrap scikit-learn implementation
        cv = sklearn.model_selection.LeaveOneGroupOut()

        for indices in cv.split(dataset, y=y, groups=groups):
            train_indices, validation_indices = indices
            test_indices: Iterable[int] = []  # no test in cross-validation
            yield train_indices, validation_indices, test_indices



from collections.abc import Iterable, Iterator, Sized
from typing import (
    Any,
    Protocol,
    runtime_checkable,
)


@runtime_checkable
class Representation(Protocol):
    def featurise(self, samples: Iterable[Any], **kwargs: Any) -> dict[str, Any]:
        """Featurises the input samples."""


@runtime_checkable
class Representations(Protocol):
    def __iter__(self) -> Iterator[Representation]: ...


@runtime_checkable
class SplittingStrategy(Protocol):
    def split(
        self,
        dataset: Sized,
        y: Iterable | None = None,
        groups: Iterable | None = None,
        **kwargs: Any,
    ) -> Iterator[tuple[Iterable[int], Iterable[int], Iterable[int]]]:
        """Generates indices to split data into training, validation, and test sets."""



"""
Tests ensuring desired API objects are part of the top-level namespace.
"""

import molflux.features


def test_exports_list_representations():
    """That the package exposes the list_representations function."""
    assert hasattr(molflux.features, "list_representations")
    assert callable(molflux.features.list_representations)


def test_exports_load_from_dict():
    """That the package exposes the load_from_dict function."""
    assert hasattr(molflux.features, "load_from_dict")
    assert callable(molflux.features.load_from_dict)


def test_exports_load_from_dicts():
    """That the package exposes the load_from_dicts function."""
    assert hasattr(molflux.features, "load_from_dicts")
    assert callable(molflux.features.load_from_dicts)


def test_exports_load_from_yaml():
    """That the package exposes the load_from_yaml function."""
    assert hasattr(molflux.features, "load_from_yaml")
    assert callable(molflux.features.load_from_yaml)


def test_exports_load_representation():
    """That the package exposes the load_representation function."""
    assert hasattr(molflux.features, "load_representation")
    assert callable(molflux.features.load_representation)


def test_exports_register_representation():
    """That the package exposes the register_representation function."""
    assert hasattr(molflux.features, "register_representation")
    assert callable(molflux.features.register_representation)


def test_exports_representation():
    """That the package exposes the Representation protocol."""
    assert hasattr(molflux.features, "Representation")


def test_exports_representations():
    """That the package exposes the Representations protocol."""
    assert hasattr(molflux.features, "Representations")



from typing import Any

try:
    from openeye import oegraphsim
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("openeye", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.openeye._utils import (
    fingerprint_to_bit_vector,
    to_oemol,
)
from molflux.features.typing import Fingerprint, MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
LINGO similarity search.

LINGO is a very fast text-based molecular similarity search method. It is based
on fragmentation of canonical isomeric SMILES strings into overlapping
substrings.
"""


class Lingo(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        **kwargs: Any,
    ) -> dict[str, list[Fingerprint]]:
        """Generates lingo fingerprints for each input molecule.

        Args:
            samples: The molecules to featurise

        Returns:
            Lingo fingerprints, as lists of bits.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('lingo')
            >>> samples = ['c1ccccc1']
            >>> representation.featurise(samples=samples)
            {'lingo': [[1, 1, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        lingo_fp_list: list[Fingerprint] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                # patch openeye support for fingerprints of empty SMILES
                if sample == "":
                    bit_vector = [0]

                else:
                    mol = to_oemol(sample)
                    fp = oegraphsim.OEFingerPrint()
                    oegraphsim.OEMakeLingoFP(fp, mol)
                    bit_vector = fingerprint_to_bit_vector(fp)

                lingo_fp_list.append(bit_vector)

        return {self.tag: lingo_fp_list}



from dataclasses import asdict
from typing import TYPE_CHECKING, Any, Literal

from pydantic.v1 import dataclasses

import datasets

if TYPE_CHECKING:
    import numpy as np

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelBase, ModelConfig
from molflux.modelzoo.models.sklearn import SKLearnClassificationMixin
from molflux.modelzoo.typing import PredictionResult
from molflux.modelzoo.utils import (
    get_concatenated_array,
    pick_features,
    validate_features,
)

try:
    from catboost import CatBoostClassifier as CBC
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("catboost", e) from e


_DESCRIPTION = """
Implementation of the scikit-learn API for a CatBoostClassifier.

CatBoost is based on gradient boosted decision trees. During training, a set
of decision trees is built consecutively. Each successive tree is built with
reduced loss compared to the previous trees.

The number of trees is controlled by the starting parameters.
To prevent overfitting, use the overfitting detector. When it is triggered,
trees stop being built.

Processing of categorical data is treated differently from other gradient tree based
methods, thus this type of model is recommended when dealing with categorical features.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
iterations : Optional[int], default: None
learning_rate: Optional[float], default: None
depth: Optional[int], default: None
l2_leaf_reg: Optional[int], default: None
model_size_reg: Optional[float], default: None
rsm: Optional[int], default: None
loss_function: Optional[Literal["Logloss", "CrossEntropy"]], default: None
border_count: Optional[int], default: None
feature_border_type: Optional[str], default: None
per_float_feature_quantization: Optional[List[str]], default: None
input_borders: Optional[str], default: None
output_borders: Optional[str], default: None
fold_permutation_block: Optional[int], default: None
od_pval: Optional[float], default: None
od_wait: Optional[int], default: None
od_type: Optional[str], default: None
nan_mode: Optional[str], default: None
counter_calc_method: Optional[str], default: None
leaf_estimation_iterations: Optional[int], default: None
leaf_estimation_method: Optional[str], default: None
thread_count: Optional[int], default: None
random_seed: Optional[int], default: None
use_best_model: Optional[bool], default: None
verbose: Optional[bool], default: None
logging_level: Optional[Union[int, str]], default: "Silent"
metric_period: Optional[int], default: None
ctr_leaf_count_limit: Optional[int], default: None
store_all_simple_ctr: Optional[bool], default: None
max_ctr_complexity: Optional[int], default: None
has_time: Optional[bool], default: None
allow_const_label: Optional[bool], default: None
classes_count: Optional[Any], default: None
class_weights: Optional[Any], default: None
one_hot_max_size: Optional[int], default: None
random_strength: Optional[int], default: None
catboost_name: Optional[str], default: None
ignored_features: Optional[Union[List[int], List[str]]], default: None
train_dir: Optional[str], default: None
custom_loss: Optional[str], default: None
custom_metric: Optional[str], default: None
eval_metric: Optional[str], default: None
bagging_temperature: Optional[float], default: None
save_snapshot: Optional[bool], default: None
snapshot_file: Optional[str], default: None
snapshot_interval: Optional[int], default: None
fold_len_multiplier: Optional[float], default: None
used_ram_limit: Optional[float], default: None
gpu_ram_part: Optional[float], default: None
allow_writing_files: Optional[bool], default: None
final_ctr_computation_mode: Optional[str], default: None
approx_on_full_history: Optional[bool], default: None
boosting_type: Optional[str], default: None
simple_ctr: Optional[bool], default: None
combinations_ctr: Optional[str], default: None
per_feature_ctr: Optional[str], default: None
task_type: Optional[str], default: None
device_config: Optional[str], default: None
devices: Optional[str], default: None
bootstrap_type: Optional[str], default: None
subsample: Optional[bool], default: None
sampling_unit: Optional[str], default: None
dev_score_calc_obj_block_size: Optional[int], default: None
max_depth: Optional[int], default: None
n_estimators: Optional[int], default: None
num_boost_round: Optional[int], default: None
num_trees: Optional[int], default: None
colsample_bylevel: Optional[int], default: None
random_state: Optional[int], default: None
reg_lambda: Optional[float], default: None
objective: Optional[str], default: None
eta: Optional[float], default: None
max_bin: Optional[int], default: None
gpu_cat_features_storage: Optional[str], default: None
data_partition: Optional[str], default: None
metadata: Optional[Dict], default: None
early_stopping_rounds: Optional[int], default: None
cat_features: Optional[List[str]], default: None
grow_policy: Optional[str], default: None
min_data_in_leaf: Optional[int], default: None
min_child_samples: Optional[int], default: None
max_leaves: Optional[int], default: None
num_leaves: Optional[int], default: None
score_function: Optional[str], default: None
leaf_estimation_backtracking: Optional[str], default: None
ctr_history_unit: Optional[int], default: None
monotone_constraints: Optional[Union[List[str], str, Dict, List]], default: None
feature_weights: Optional[Union[List, str, Dict]], default: None
penalties_coefficient: Optional[float], default: None
first_feature_use_penalties: Optional[bool], default: None
model_shrink_rate: Optional[float], default: None
model_shrink_mode: Optional[str], default: None
langevin: Optional[bool], default: None
diffusion_temperature: Optional[float], default: None
posterior_sampling: Optional[bool], default: None
boost_from_average: Optional[bool], default: None
dictionaries: Optional[Any], default: None
feature_calcers: Optional[Any], default: None
scale_pos_weight: Optional[Any], default: None
text_features: Optional[Any], default: None
text_processing: Optional[Any], default: None
tokenizers: Optional[Any], default: None
"""

LossFunctions = Literal["Logloss", "CrossEntropy"]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class CatBoostClassifierConfig(ModelConfig):
    iterations: int | None = None
    learning_rate: float | None = None
    depth: int | None = None
    l2_leaf_reg: int | None = None
    model_size_reg: float | None = None
    rsm: int | None = None
    loss_function: LossFunctions | None = None
    border_count: int | None = None
    feature_border_type: str | None = None
    per_float_feature_quantization: list[str] | None = None
    input_borders: str | None = None
    output_borders: str | None = None
    fold_permutation_block: int | None = None
    od_pval: float | None = None
    od_wait: int | None = None
    od_type: str | None = None
    nan_mode: str | None = None
    counter_calc_method: str | None = None
    leaf_estimation_iterations: int | None = None
    leaf_estimation_method: str | None = None
    thread_count: int | None = None
    random_seed: int | None = None
    use_best_model: bool | None = None
    verbose: bool | None = None
    logging_level: int | str | None = "Silent"
    metric_period: int | None = None
    ctr_leaf_count_limit: int | None = None
    store_all_simple_ctr: bool | None = None
    max_ctr_complexity: int | None = None
    has_time: bool | None = None
    allow_const_label: bool | None = None
    one_hot_max_size: int | None = None
    random_strength: int | None = None
    catboost_name: str | None = None
    ignored_features: list[int] | list[str] | None = None
    train_dir: str | None = None
    custom_metric: str | None = None
    eval_metric: str | None = None
    bagging_temperature: float | None = None
    save_snapshot: bool | None = None
    snapshot_file: str | None = None
    snapshot_interval: int | None = None
    fold_len_multiplier: float | None = None
    used_ram_limit: float | None = None
    gpu_ram_part: float | None = None
    allow_writing_files: bool | None = None
    final_ctr_computation_mode: str | None = None
    approx_on_full_history: bool | None = None
    boosting_type: str | None = None
    simple_ctr: bool | None = None
    combinations_ctr: str | None = None
    per_feature_ctr: str | None = None
    task_type: str | None = None
    device_config: str | None = None
    devices: str | None = None
    bootstrap_type: str | None = None
    subsample: bool | None = None
    sampling_unit: str | None = None
    dev_score_calc_obj_block_size: int | None = None
    max_depth: int | None = None
    n_estimators: int | None = None
    num_boost_round: int | None = None
    num_trees: int | None = None
    colsample_bylevel: int | None = None
    random_state: int | None = None
    reg_lambda: float | None = None
    objective: str | None = None
    eta: float | None = None
    max_bin: int | None = None
    gpu_cat_features_storage: str | None = None
    data_partition: str | None = None
    metadata: dict | None = None
    early_stopping_rounds: int | None = None
    cat_features: list[str] | None = None
    grow_policy: str | None = None
    min_data_in_leaf: int | None = None
    min_child_samples: int | None = None
    max_leaves: int | None = None
    num_leaves: int | None = None
    score_function: str | None = None
    leaf_estimation_backtracking: str | None = None
    ctr_history_unit: int | None = None
    monotone_constraints: list[str] | str | dict | list | None = None
    feature_weights: list | str | dict | None = None
    penalties_coefficient: float | None = None
    first_feature_use_penalties: bool | None = None
    model_shrink_rate: float | None = None
    model_shrink_mode: str | None = None
    langevin: bool | None = None
    diffusion_temperature: float | None = None
    posterior_sampling: bool | None = None
    boost_from_average: bool | None = None
    auto_class_weights: str | None = None
    dictionaries: Any | None = None
    feature_calcers: Any | None = None
    scale_pos_weight: Any | None = None
    text_features: Any | None = None
    text_processing: Any | None = None
    tokenizers: Any | None = None


class CatBoostClassifier(
    SKLearnClassificationMixin,
    ModelBase[CatBoostClassifierConfig],
):
    """
    Note:
        The cat_features must specify the column names of the input Dataset object. If
        the column is a column of arrays (for example a column of fingerprints), it
        is assumed that each array molflux is a categorical column of its own.
    """

    @property
    def config(self) -> dict[str, Any]:
        return asdict(self.model_config)

    @property
    def _config_builder(self) -> type[CatBoostClassifierConfig]:
        return CatBoostClassifierConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> CBC:
        """Note: Not yet pulling the config in"""
        config: CatBoostClassifierConfig = self.model_config
        return CBC(
            iterations=config.iterations,
            learning_rate=config.learning_rate,
            depth=config.depth,
            l2_leaf_reg=config.l2_leaf_reg,
            model_size_reg=config.model_size_reg,
            rsm=config.rsm,
            loss_function=config.loss_function,
            border_count=config.border_count,
            feature_border_type=config.feature_border_type,
            per_float_feature_quantization=config.per_float_feature_quantization,
            input_borders=config.input_borders,
            output_borders=config.output_borders,
            fold_permutation_block=config.fold_permutation_block,
            od_pval=config.od_pval,
            od_wait=config.od_wait,
            od_type=config.od_type,
            nan_mode=config.nan_mode,
            counter_calc_method=config.counter_calc_method,
            leaf_estimation_iterations=config.leaf_estimation_iterations,
            leaf_estimation_method=config.leaf_estimation_method,
            thread_count=config.thread_count,
            random_seed=config.random_seed,
            use_best_model=config.use_best_model,
            verbose=config.verbose,
            logging_level=config.logging_level,
            metric_period=config.metric_period,
            ctr_leaf_count_limit=config.ctr_leaf_count_limit,
            store_all_simple_ctr=config.store_all_simple_ctr,
            max_ctr_complexity=config.max_ctr_complexity,
            has_time=config.has_time,
            allow_const_label=config.allow_const_label,
            one_hot_max_size=config.one_hot_max_size,
            random_strength=config.random_strength,
            name=config.catboost_name,
            ignored_features=config.ignored_features,
            train_dir=config.train_dir,
            custom_metric=config.custom_metric,
            eval_metric=config.eval_metric,
            bagging_temperature=config.bagging_temperature,
            save_snapshot=config.save_snapshot,
            snapshot_file=config.snapshot_file,
            snapshot_interval=config.snapshot_interval,
            fold_len_multiplier=config.fold_len_multiplier,
            used_ram_limit=config.used_ram_limit,
            gpu_ram_part=config.gpu_ram_part,
            allow_writing_files=config.allow_writing_files,
            final_ctr_computation_mode=config.final_ctr_computation_mode,
            approx_on_full_history=config.approx_on_full_history,
            boosting_type=config.boosting_type,
            simple_ctr=config.simple_ctr,
            combinations_ctr=config.combinations_ctr,
            per_feature_ctr=config.per_feature_ctr,
            task_type=config.task_type,
            device_config=config.device_config,
            devices=config.devices,
            bootstrap_type=config.bootstrap_type,
            subsample=config.subsample,
            sampling_unit=config.sampling_unit,
            dev_score_calc_obj_block_size=config.dev_score_calc_obj_block_size,
            max_depth=config.max_depth,
            n_estimators=config.n_estimators,
            num_boost_round=config.num_boost_round,
            num_trees=config.num_trees,
            colsample_bylevel=config.colsample_bylevel,
            random_state=config.random_state,
            reg_lambda=config.reg_lambda,
            objective=config.objective,
            eta=config.eta,
            max_bin=config.max_bin,
            gpu_cat_features_storage=config.gpu_cat_features_storage,
            data_partition=config.data_partition,
            metadata=config.metadata,
            early_stopping_rounds=config.early_stopping_rounds,
            cat_features=config.cat_features,
            grow_policy=config.grow_policy,
            min_data_in_leaf=config.min_data_in_leaf,
            min_child_samples=config.min_child_samples,
            max_leaves=config.max_leaves,
            num_leaves=config.num_leaves,
            score_function=config.score_function,
            leaf_estimation_backtracking=config.leaf_estimation_backtracking,
            ctr_history_unit=config.ctr_history_unit,
            monotone_constraints=config.monotone_constraints,
            feature_weights=config.feature_weights,
            penalties_coefficient=config.penalties_coefficient,
            first_feature_use_penalties=config.first_feature_use_penalties,
            model_shrink_rate=config.model_shrink_rate,
            model_shrink_mode=config.model_shrink_mode,
            langevin=config.langevin,
            diffusion_temperature=config.diffusion_temperature,
            posterior_sampling=config.posterior_sampling,
            boost_from_average=config.boost_from_average,
            auto_class_weights=config.auto_class_weights,
            dictionaries=config.dictionaries,
            feature_calcers=config.feature_calcers,
            scale_pos_weight=config.scale_pos_weight,
            text_features=config.text_features,
            text_processing=config.text_processing,
            tokenizers=config.tokenizers,
        )

    def _train(
        self,
        train_data: datasets.Dataset,
        **kwargs: Any,
    ) -> Any:
        """
        Note:
            This is not inheriting from SKLearn because of the special data treatment
        that needs to be done for categorical feature handling
        """

        # validate y features as well
        validate_features(train_data, self.y_features)

        x_data = pick_features(train_data, self.x_features)
        X = get_concatenated_array(x_data, self.x_features)

        y_data = pick_features(train_data, self.y_features)
        y = get_concatenated_array(y_data, self.y_features)

        # instantiate model
        self.model: CBC = self._instantiate_model()

        # train
        self.model.fit(X, y)

    def _predict(
        self,
        data: datasets.Dataset,
        **kwargs: Any,
    ) -> PredictionResult:
        """Note: same as _train, see above"""

        # TODO(avianello): for now catboost does not support multitask
        display_names = self._predict_display_names
        display_name = display_names[0]

        if not len(data):
            return {display_name: []}

        X = get_concatenated_array(data, self.x_features)
        y_predict: np.ndarray = self.model.predict(X)

        return {display_name: y_predict.tolist()}



import numpy as np
import pytest

from molflux.splits.catalogue import list_splitting_strategies
from molflux.splits.load import load_splitting_strategy
from molflux.splits.strategy import SplittingStrategy

strategy_name = "group_k_fold"


@pytest.fixture(scope="module")
def fixture_test_strategy():
    return load_splitting_strategy(strategy_name)


@pytest.fixture(scope="module")
def fixture_sample_inputs():
    n = 100
    dataset = np.random.rand(n)
    groups = np.random.choice(["a", "b", "c"], size=n)
    return dataset, groups


def test_is_in_catalogue():
    """That the strategy is registered in the catalogue."""
    catalogue = list_splitting_strategies()
    all_strategy_names = [name for names in catalogue.values() for name in names]
    assert strategy_name in all_strategy_names


def test_implements_protocol(fixture_test_strategy):
    """That the strategy implements the protocol."""
    strategy = fixture_test_strategy
    assert isinstance(strategy, SplittingStrategy)


def test_requires_groups(fixture_test_strategy):
    """That you need to provide groups to use this strategy."""
    dataset = [0, 1, 2, 3, 4, 5]
    strategy = fixture_test_strategy
    with pytest.raises(ValueError):
        next(strategy.split(dataset))


def test_yields_more_than_one_fold(fixture_sample_inputs, fixture_test_strategy):
    """That the splitting strategy yields several folds."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, groups=groups)
    assert len(list(indices)) > 1


def test_cannot_set_only_one_fold(fixture_sample_inputs, fixture_test_strategy):
    """That cannot perform cross validation with only one fold."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    with pytest.raises(ValueError):
        next(strategy.split(dataset=dataset, groups=groups, n_splits=1))


def test_more_splits_than_groups_raises(fixture_test_strategy):
    """That cannot have more splits than number of unique groups."""
    dataset = [0, 1, 2, 3, 4, 5]
    groups = ["a", "b", "a", "a", "b", "c"]
    strategy = fixture_test_strategy
    with pytest.raises(ValueError):
        next(strategy.split(dataset, groups=groups, n_splits=4))


def test_splits_are_disjoint(fixture_sample_inputs, fixture_test_strategy):
    """That data is spread across splits without overlap."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = strategy.split(dataset=dataset, groups=groups)
    for train_indices, validation_indices, _ in indices:
        assert set(train_indices).isdisjoint(set(validation_indices))


def test_each_sample_in_a_different_fold(fixture_sample_inputs, fixture_test_strategy):
    """That each sample ends up in a different fold."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = strategy.split(dataset, groups=groups, n_splits=2)
    for train_indices, validation_indices, _ in indices:
        groups_in_train = set(groups[train_indices])
        groups_in_validation = set(groups[validation_indices])
        assert set(groups_in_train).isdisjoint(set(groups_in_validation))


def test_deterministic(fixture_sample_inputs, fixture_test_strategy):
    """That repeated splits are deterministic."""
    dataset, groups = fixture_sample_inputs
    strategy = fixture_test_strategy

    # Do one iteration of k-fold splits
    indices_a = strategy.split(dataset, groups=groups, n_splits=2)
    train_indices_a1, validation_indices_a1, _ = next(indices_a)
    train_indices_a2, validation_indices_a2, _ = next(indices_a)

    # And again...
    indices_b = strategy.split(dataset, groups=groups, n_splits=2)
    train_indices_b1, validation_indices_b1, _ = next(indices_b)
    train_indices_b2, validation_indices_b2, _ = next(indices_b)

    # Check that we get the same results across folds
    assert all(i == j for i, j in zip(train_indices_a1, train_indices_b1, strict=False))
    assert all(
        i == j
        for i, j in zip(validation_indices_a1, validation_indices_b1, strict=False)
    )
    assert all(i == j for i, j in zip(train_indices_a2, train_indices_b2, strict=False))
    assert all(
        i == j
        for i, j in zip(validation_indices_a2, validation_indices_b2, strict=False)
    )



from collections.abc import Iterable
from typing import Any

from molflux.features.catalogue import get_representation_cls
from molflux.features.parsers import Spec, dict_parser, yaml_parser
from molflux.features.representation import Representation, Representations
from molflux.features.typing import PathLike


def load_representation(name: str, **init_kwargs: Any) -> Representation:
    """Loads a `Representation` from the catalogue.

    Examples:
        >>> from molflux.features import load_representation
        >>> representation = load_representation("canonical_smiles")
    """

    # Fetch relevant representation class from the catalogue
    representation_cls = get_representation_cls(representation_name=name)

    # Instantiate representation object
    representation = representation_cls(**init_kwargs)

    return representation


def _load_from_spec(spec: Spec) -> Representation:
    """Loads a representation from a validated Spec."""

    # Build representation
    representation = load_representation(name=spec.name, **spec.config)

    # update state
    representation.update_state(**spec.presets)

    return representation


def load_from_dict(dictionary: dict[str, Any]) -> Representation:
    """Loads a representation from a config dict."""

    # Validate dictionary
    spec = dict_parser(dictionary=dictionary)

    return _load_from_spec(spec=spec)


def load_from_dicts(dictionaries: Iterable[dict[str, Any]]) -> Representations:
    """Loads a collection of Representations from an iterable of dicts."""

    representations = Representations()
    for dictionary in dictionaries:
        representation = load_from_dict(dictionary=dictionary)

        representations.add_representation(representation=representation)

    return representations


def load_from_yaml(path: PathLike) -> Representations:
    """Loads a collection of representations from a yaml config file."""

    specs = yaml_parser(path=path)

    representations = Representations()
    for spec in specs:
        representation = _load_from_spec(spec=spec)

        representations.add_representation(representation=representation)

    return representations



from typing import Any

import pandas
import pyarrow.interchange

import datasets


def hf_dataset_from_dataframe(df: Any) -> datasets.Dataset:
    """Builds a huggingface datasets.Dataset from any DataFrame supporting the
    DataFrame Interchange Protocol.

    Huggingface datasets do not support the interchange protocol quite yet.
    We therefore first interchange the input dataframe into a pyarrow table
    and then build a dataset from it.

    References:
        https://data-apis.org/dataframe-protocol/latest/purpose_and_scope.html
    """

    if isinstance(df, datasets.Dataset):
        return df

    # pandas DataFrames are handled separately as they are not always compatible
    # with the pyarrow.interchange call down below
    if isinstance(df, pandas.DataFrame):
        return datasets.Dataset.from_pandas(df)

    try:
        return datasets.Dataset(pyarrow.interchange.from_dataframe(df))
    except NotImplementedError as e:
        raise TypeError(
            f"Could not interchange input dataframe of type {type(df)!r} into an arrow-backed dataframe object. Most likely, one of the columnar data types is not supported by Apache Arrow (pyarrow) backends.",
        ) from e



from typing import Any

try:
    from openeye import oegraphsim
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("openeye", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.openeye._utils import (
    fingerprint_to_bit_vector,
    to_oemol,
)
from molflux.features.typing import Fingerprint, MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
A circular fingerprint is generated by exhaustively enumerating all circular
fragments grown radially from each heavy atom of the molecule up to the given
radius and then hashing these fragments into a fixed-length bitvector
"""


class Circular(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        length: int = 2048,
        diameter: int = 6,
        atom_type: int = oegraphsim.OEFPAtomType_DefaultAtom,
        bond_type: int = oegraphsim.OEFPBondType_DefaultBond,
        **kwargs: Any,
    ) -> dict[str, list[Fingerprint]]:
        """Generates a circular fingerprint for each input molecule.

        Args:
            samples: The molecules to featurise.
            length: The size of the fingerprint in bits. This number has to be
                larger than or equal to 2^4 and smaller than 2^16.
            diameter: The largest circular fragments (in diameter) that are
                enumerated during the fingerprint generation. All enumerated
                circular fragments are hashed into the OEFingerPrint object.
            atom_type: Defines which atom properties are encoded during the
                fingerprint generation. This value has to be either a value or
                a set of bitwise OR`d values from the `OEFPAtomType` namespace.
            bond_type: Defines which bond properties are encoded during the
                fingerprint generation. This value has to be either a value or
                a set of bitwise OR`d values from the `OEFPBondType` namespace.

        Returns:
            Circular fingerprints, as lists of bits.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('circular')
            >>> samples = ['c1ccccc1']
            >>> representation.featurise(samples=samples, length=16)
            {'circular': [[0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0]]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        if not ((length & (length - 1) == 0) and length != 0):
            raise RuntimeError(f"length: {length} must be a power of 2")

        if (length < 2**4) or (length >= 2**16):
            raise ValueError(
                "The fingerprint size has to be larger than or equal to 2^4 and smaller than 2^16",
            )

        if not ((diameter % 2 == 0) and diameter >= 0):
            raise RuntimeError(f"diameter: {diameter} must be even and >= 0")

        circular_fp_list: list[Fingerprint] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                # patch openeye support for fingerprints of empty SMILES
                if sample == "":
                    bit_vector = [0] * length

                else:
                    mol = to_oemol(sample)
                    fp = oegraphsim.OEFingerPrint()
                    oegraphsim.OEMakeCircularFP(
                        fp,
                        mol,
                        length,
                        0,
                        int(diameter / 2),
                        atom_type,
                        bond_type,
                    )
                    bit_vector = fingerprint_to_bit_vector(fp)

                circular_fp_list.append(bit_vector)

        return {self.tag: circular_fp_list}



import json
import os
import pathlib
import warnings
from typing import Any, Literal, cast, get_args

import fsspec

import datasets
from datasets import Dataset, DatasetDict, IterableDataset, IterableDatasetDict
from datasets.filesystems import is_remote_filesystem
from datasets.utils.py_utils import NestedDataStructure
from molflux.datasets.typing import DataFiles, FileSystem, HFDataFiles, PathLike
from molflux.datasets.utils import (
    is_cloud_data,
    sanitise_data_files_as_str,
)

SupportedFileFormats = Literal["disk", "parquet", "csv", "json"]
_DEFAULT_DATASET_DICT_FORMAT: SupportedFileFormats = "parquet"


def __resolve_path_file_format(path: str) -> SupportedFileFormats:
    """Resolves the core file format of persisted assets.

    This function parses file paths to resolve them into one of our
    supported file formats. Directory-like paths are resolved to 'disk' format.
    Compressed file formats are resolved to their core uncompressed file format.
    """

    suffixes = [suffix.lstrip(".") for suffix in pathlib.Path(path).suffixes]
    is_directory = not suffixes

    if is_directory:
        return cast(SupportedFileFormats, "disk")  # cast to help type checker

    for supported_file_format in get_args(SupportedFileFormats):
        if supported_file_format in suffixes:
            return cast(
                SupportedFileFormats,
                supported_file_format,
            )  # cast to help type checker

    raise ValueError(
        f"Unsupported dataset format for path {path!r}: expected one of {get_args(SupportedFileFormats)!r}",
    )


def _resolve_data_files_file_format(data_files: HFDataFiles) -> SupportedFileFormats:
    """Resolves the file format of the input huggingface DataFiles.

    All data files are expected to resolve to the same file format.
    """
    file_formats: list[SupportedFileFormats] = []
    for path in NestedDataStructure(data_files).flatten():
        file_format = __resolve_path_file_format(path)
        file_formats.append(file_format)

    if not len(set(file_formats)) == 1:
        raise ValueError("Mismatched file formats.")

    return file_formats[0]


def __raise_for_path_format_consistency(
    path: str,
    format: str,
    is_dataset_dict: bool,
) -> None:
    """Checks that the given path / format combination is valid.

    Args:
        path: The target path to save to / load from the dataset.
        format: The persistence format the dataset should be in / is.
        is_dataset_dict: Whether the target of the IO operation is / was a DatasetDict.
    """

    # only specific persistence formats are supported
    if format not in get_args(SupportedFileFormats):
        raise ValueError(
            f"Unsupported dataset format for path {path!r}: got {format!r} expected one of {get_args(SupportedFileFormats)!r}",
        )

    suffix = pathlib.Path(path).suffix
    is_directory_like_path = not suffix

    # DatasetDicts can only be saved / loaded from directories
    if is_dataset_dict and not is_directory_like_path:
        raise ValueError(
            f"Invalid path for DatasetDict: {path!r} is not a directory-like path",
        )

    # For Datasets, the filepath extension must match the format
    if not is_dataset_dict:
        expected_format = _resolve_data_files_file_format(path)
        if format != expected_format:
            raise ValueError(
                f"Mismatched file format: got {format!r} expected {expected_format!r}",
            )


def _raise_for_data_files_format_consistency(
    data_files: HFDataFiles,
    format: str,
    is_dataset_dict: bool,
) -> None:
    """Checks that the given data_files / format combination is valid."""
    for path in NestedDataStructure(data_files).flatten():
        __raise_for_path_format_consistency(
            path,
            format=format,
            is_dataset_dict=is_dataset_dict,
        )


def _check_compatible_filesystem(fs: FileSystem, data_files: HFDataFiles) -> None:
    """Checks that the given filesystem is compatible with the data files it
    should operate on.

    Cloud data files should be backed by a cloud filesystem, local data files
    should be backed by a local filesystem.
    """
    need_to_load_from_cloud = is_cloud_data(data_files)
    is_cloud_filesystem = is_remote_filesystem(fs)
    if need_to_load_from_cloud and not is_cloud_filesystem:
        raise ValueError(
            "Incompatible filesystem: please provide a remote filesystem for accessing cloud data.",
        )
    if not need_to_load_from_cloud and is_cloud_filesystem:
        raise ValueError(
            "Incompatible filesystem: please provide a local filesystem for accessing local data.",
        )


def _resolve_file_system_for_data_files(data_files: HFDataFiles) -> FileSystem:
    """Returns a ready-made file system suited for handling input data files."""
    need_to_load_from_cloud = is_cloud_data(data_files)
    fs = (
        fsspec.filesystem("s3")
        if need_to_load_from_cloud
        else fsspec.filesystem("file")
    )
    return fs


def save_dataset_to_store(
    dataset: Dataset | DatasetDict,
    path: PathLike,
    format: SupportedFileFormats | None = None,
    fs: FileSystem | None = None,
    **kwargs: Any,
) -> None:
    """Save a dataset to persistent storage.

    Args:
        dataset: The Dataset or DatasetDict to persist.
        path: The target path where to save the dataset. Should be a directory
            for DatasetDicts or for saving datasets as 'disk' format.
        format: The file format to save the dataset as. Should be one of
            ["disk", "parquet", "csv", "json"]. If missing, for Datasets this
            will be automatically inferred from 'path'. For DatasetDicts, it
            will default to "parquet".
        fs: The filesystem object to use to save the dataset. If not provided,
            an appropriate filesystem will be automatically provided for
            saving the data.

    Examples:
        >>> from molflux.datasets import save_dataset_to_store  # doctest: +SKIP
        # save a Dataset as parquet
        >>> save_dataset_to_store(dataset, "s3://my/dataset/data.parquet")  # doctest: +SKIP
        # save a Dataset as a bunch of .arrow files
        >>> save_dataset_to_store(dataset, "s3://my/dataset/data")  # doctest: +SKIP
        # save a DatasetDict as parquet
        >>> save_dataset_to_store(dataset_dict, "s3://my/dataset/data")  # doctest: +SKIP
        # save a DatasetDict as a specific persistence format (e.g. csv)
        >>> save_dataset_to_store(dataset_dict, "s3://my/dataset/data", format="csv")  # doctest: +SKIP
        # save a DatasetDict as a bunch of .arrow files
        >>> save_dataset_to_store(dataset_dict, "s3://my/dataset/data", format="disk")  # doctest: +SKIP
    """

    path = str(path)

    # Provide a filesystem if one not specified (convenience)
    if fs is None:
        fs = _resolve_file_system_for_data_files(data_files=path)
    _check_compatible_filesystem(fs, data_files=path)

    # Resolve format if one not specified (convenience)
    is_dataset_dict = isinstance(dataset, datasets.DatasetDict)
    if format is None:
        format = (
            _DEFAULT_DATASET_DICT_FORMAT
            if is_dataset_dict
            else _resolve_data_files_file_format(path)
        )

    _raise_for_data_files_format_consistency(
        path,
        format=format,
        is_dataset_dict=is_dataset_dict,
    )

    # save_to_disk is implemented both for Dataset and DatasetDict
    if format == "disk":
        return dataset.save_to_disk(path, storage_options=fs.storage_options, **kwargs)  # type: ignore[no-any-return]

    # for all other file format let's implement our own wrappers
    if is_dataset_dict:
        return _save_dataset_dict_to_file(
            dataset_dict=dataset,
            dataset_dict_path=path,
            format=format,
            fs=fs,
            **kwargs,
        )
    else:
        return _save_dataset_to_file(
            dataset=dataset,
            dataset_path=path,
            format=format,
            fs=fs,
            **kwargs,
        )


def _generate_split_path(root: str, split: str, format: SupportedFileFormats) -> str:
    """Generates a standardised name for filenames of splits."""
    return os.path.join(root, f"{split}.{format}")


def _save_dataset_dict_to_file(
    dataset_dict: DatasetDict,
    dataset_dict_path: str,
    format: SupportedFileFormats,
    fs: FileSystem,
    **kwargs: Any,
) -> None:
    """Saves a DatasetDict to file.

    This emulates what huggingface already implements out of the box for
    DatasetDict.save_to_disk(), extending it to all other file formats too. A
    json metadata file is saved with the names of the splits composing the
    DatasetDict, and the individual datasets in the dict are saved individually
    alongside of it.
    """

    is_local = not is_remote_filesystem(fs)
    if is_local:
        pathlib.Path(dataset_dict_path).resolve().mkdir(parents=True, exist_ok=True)

    # write out huggingface's dataset_dict.json metadata file
    dataset_dict_json_filepath = os.path.join(
        dataset_dict_path,
        datasets.config.DATASETDICT_JSON_FILENAME,
    )
    with fs.open(dataset_dict_json_filepath, "w", encoding="utf-8") as f:
        json.dump({"splits": list(dataset_dict)}, f)

    # write out individual datasets
    for k, dataset in dataset_dict.items():
        dataset_path = _generate_split_path(dataset_dict_path, split=k, format=format)
        _save_dataset_to_file(
            dataset,
            dataset_path=dataset_path,
            format=format,
            fs=fs,
            **kwargs,
        )


def _save_dataset_to_file(
    dataset: Dataset,
    dataset_path: str,
    format: SupportedFileFormats,
    fs: FileSystem,
    **kwargs: Any,
) -> None:
    """Saves a single Dataset to file."""

    is_local = not is_remote_filesystem(fs)
    if is_local:
        pathlib.Path(dataset_path).resolve().parent.mkdir(parents=True, exist_ok=True)

    if not fs.isfile(dataset_path):
        fs.touch(dataset_path)

    with fs.open(dataset_path, "wb") as f:
        if format == "csv":
            dataset.to_csv(path_or_buf=f, **kwargs)

        elif format == "parquet":
            dataset.to_parquet(path_or_buf=f, **kwargs)

        elif format == "json":
            dataset.to_json(path_or_buf=f, **kwargs)

        else:
            raise NotImplementedError(f"Invalid file format: {format}")


def _is_dataset_dict_dir(dest_dataset_dict_path: str, fs: FileSystem) -> bool:
    """Checks that we are in a dataset_dict artefact directory.

    This is done by checking for the presence of the json metadata file that
    gets generated on saving of DatasetDicts. This emulates the same logic
    already implemented natively in huggingface datasets.load_from_disk(), but
    extending it for all other formats too.
    """

    if not fs.isdir(dest_dataset_dict_path):
        return False

    dataset_dict_json_path = os.path.join(
        dest_dataset_dict_path,
        datasets.config.DATASETDICT_JSON_FILENAME,
    )
    return fs.isfile(dataset_dict_json_path)  # type: ignore[no-any-return]


def _resolve_expected_dataset_dict_dir_data_files(
    dest_dataset_dict_path: str,
    format: SupportedFileFormats,
    fs: FileSystem,
) -> dict[str, str]:
    """Expands the path to a DatasetDict directory into its expected DataFiles.

    This works by extracting the split names from the persisted DatasetDict json
    metadata file and then using those to resolve the expected dataset filenames.

    Examples:
        >>> _resolve_expected_dataset_dict_dir_data_files("my/data", format="parquet")  # doctest: +SKIP
        {"train": "my/data/train.parquet", "validation": "my/data/validation.parquet", "test": "my/data/test.parquet"}
    """

    if not _is_dataset_dict_dir(dest_dataset_dict_path, fs=fs):
        raise FileNotFoundError(
            "Expected to load a `DatasetDict` object, but provided path is not a `DatasetDict`.",
        )

    dataset_dict_json_path = os.path.join(
        dest_dataset_dict_path,
        datasets.config.DATASETDICT_JSON_FILENAME,
    )
    with fs.open(dataset_dict_json_path, "r", encoding="utf-8") as f:
        splits = json.load(f)["splits"]

    return {
        k: _generate_split_path(dest_dataset_dict_path, split=k, format=format)
        for k in splits
    }


def load_dataset_from_store(
    source: DataFiles,
    format: SupportedFileFormats | None = None,
    fs: FileSystem | None = None,
    split: str | None = None,
    **kwargs: Any,
) -> Dataset | DatasetDict | IterableDataset | IterableDatasetDict:
    """Loads a dataset from persisted storage.

    Args:
        source: Path(s) to source data file(s).
        format: The file format the dataset has been persisted as. Should be one of
            ["disk", "parquet", "csv", "json"]. If missing, for Datasets this
            will be automatically inferred from 'source'. For DatasetDicts, it
            will default to "parquet".
        fs: The filesystem object to use to save the dataset. If not provided,
            an appropriate filesystem will be automatically provided for
            saving the data.
        split: For DatasetDicts, which split of the data to load. If not provided,
            defaults to None which will return a DatasetDict of all splits. Any
            other value will return a single Dataset corresponding to the indicated
            split. The special value "all" can be used to return a single Dataset
            with all splits merged.

    Examples:
        >>> from molflux.datasets import load_dataset_from_store  # doctest: +SKIP
        # load a persisted parquet Dataset
        >>> dataset = load_dataset_from_store("s3://my/bucket/data.parquet")  # doctest: +SKIP
        # load a persisted 'disk' Dataset
        >>> dataset = load_dataset_from_store("s3://my/bucket/data")  # doctest: +SKIP
        # load a persisted parquet DatasetDict
        >>> dataset_dict = load_dataset_from_store("s3://my/bucket/data")  # doctest: +SKIP
        # load a persisted csv DatasetDict
        >>> dataset_dict = load_dataset_from_store("s3://my/bucket/data", format="csv")  # doctest: +SKIP
        # load a persisted parquet DatasetDict as a single Dataset
        >>> dataset = load_dataset_from_store("s3://my/bucket/data", split="all")  # doctest: +SKIP
        # load from a mapping of persisted Datasets
        >>> source = {"train": "my/train_data.csv", "test": "my_test_data.csv"}
        >>> dataset_dict = load_dataset_from_store(source)  # doctest: +SKIP
    """

    # Collapse path-like datafiles to str
    source = sanitise_data_files_as_str(source)

    # Provide a filesystem if one not specified (convenience)
    if fs is None:
        fs = _resolve_file_system_for_data_files(data_files=source)
    _check_compatible_filesystem(fs, data_files=source)

    # Resolve format if one not specified (convenience)
    is_dataset_dict = isinstance(source, str) and _is_dataset_dict_dir(source, fs=fs)
    if format is None:
        format = (
            _DEFAULT_DATASET_DICT_FORMAT
            if is_dataset_dict
            else _resolve_data_files_file_format(source)
        )

        _raise_for_data_files_format_consistency(
            source,
            format=format,
            is_dataset_dict=is_dataset_dict,
        )

    # huggingface implements their own dedicated function for loading "disk" files
    if format == "disk":
        # Loading from disk only allows loading from a single directory
        if not isinstance(source, str) or not fs.isdir(source):
            raise ValueError(
                f"Invalid data source: {source} should point to a (existing) directory for format {format!r}.",
            )

        return datasets.load_from_disk(
            dataset_path=source,
            storage_options=fs.storage_options,
        )

    # for all other file format let's implement our own wrapper:

    # Auto-expand directory source into datadict datafiles (convenience)
    if is_dataset_dict:
        data_files = _resolve_expected_dataset_dict_dir_data_files(
            source,  # type: ignore[arg-type]
            format=format,
            fs=fs,
        )
    else:
        data_files = source  # type: ignore[assignment]

    # Change hf default behaviour and return single dataset as Dataset (convenience)
    data_files_are_for_dataset_dict = is_dataset_dict or isinstance(data_files, dict)
    if split is None and not data_files_are_for_dataset_dict:
        split = "all"

    with warnings.catch_warnings():
        # Do not pass through unhandled deprecation warnings from dependencies of hf datasets themselves
        warnings.filterwarnings(
            "ignore",
            message="The .* keyword in pd.read_csv is deprecated and will be removed in a future version",
            category=FutureWarning,
            module="datasets[.*]",
        )

        return datasets.load_dataset(
            format,
            data_files=data_files,
            split=split,
            storage_options=fs.storage_options,
            **kwargs,
        )



from collections.abc import Iterator
from typing import Any

from molflux.splits.bases import SplittingStrategyBase
from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.strategies.core._ordered_split_utils import (
    find_begin_end_indices,
    force_min_test_size,
    force_same_set_same_group,
    make_sorted_groups_indices,
    validate_ordered_strategy_arguments,
)
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable

_DESCRIPTION = """
A split based on the ordering of rows defined by the group column. This can be used for any sort of temporal splits.
Effectively, this is a sorting of the dataset by the groups column, and then a linear split.
"""


class OrderedSplit(SplittingStrategyBase):
    def _info(self) -> SplittingStrategyInfo:
        return SplittingStrategyInfo(
            description=_DESCRIPTION,
        )

    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        *,
        train_fraction: float | list[float] = 0.8,
        gap_train_validation_fraction: float | list[float] | None = None,
        validation_fraction: float | list[float] = 0.1,
        gap_validation_test_fraction: float | list[float] | None = None,
        test_fraction: float | list[float] = 0.1,
        gap_test_end_fraction: float | list[float] | None = None,
        undefined_groups_in_train: bool = True,
        min_test_size: int | list[int] | None = None,
        same_group_same_set: bool = True,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Args:
            dataset: The data to be split.
            groups: Group labels for the samples used while splitting the dataset.
            train_fraction (float or list): The proportion of the dataset to include in the train split.
                If a list is specified, split returns multiple folds with the specified fraction.
                Might not always be possible to get the exact fraction.
                Defaults to 0.8.
            gap_train_validation_fraction (Optional): Fraction of data to leave out between train and validation.
                Must be a float or a list, same as the train, validation fractions.
                Defaults to None
            validation_fraction: The proportion of the dataset to include in the validation split.
                If a list is specified, split returns multiple folds with the specified fraction.
                Might not always be possible to get the exact fraction.
                Defaults to 0.1.
            gap_validation_test_fraction (Optional): Fraction of data to leave out between validation and test.
                Must be a float or a list, same as the validation, test fractions.
                Defaults to None
            test_fraction: The proportion of the dataset to include in the test split.
                If a list is specified, split returns multiple folds with the specified fraction.
                Might not always be possible to get the exact fraction.
                Defaults to 0.1.
            gap_test_end_fraction (Optional): Fraction of data to leave out after test ste.
                Must be a float or a list, same as the test fractions.
                Defaults to None
            undefined_groups_in_train: If True, any row with no group (None) will be added to the train set (split ratio
                might be different from specified). Defaults to True.
            min_test_size: The minimum number of datapoints in the test set. Overrides the ratios.
                Can be an int for all the folds, or a list of ints with the same length as the folds. Defaults to None.
            same_group_same_set: Forces datapoints with the same group to be in the same set. Defaults to True.

            Examples:
            >>> from molflux.splits import load_splitting_strategy
            >>> strategy = load_splitting_strategy("ordered_split")
            >>>
            >>> x = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            >>> groups = [0, 0, 0, 2, 2, 3, 1, 1, 1, 1]
            >>>
            >>> next(strategy.split(x, groups=groups))
            ([0, 1, 2, 6, 7, 8, 9], [3, 4], [5])
        """

        # verify that all arguments are valid (and massage them into standard form
        validated_dict = validate_ordered_strategy_arguments(
            groups=groups,
            train_fraction=train_fraction,
            gap_train_validation_fraction=gap_train_validation_fraction,
            validation_fraction=validation_fraction,
            gap_validation_test_fraction=gap_validation_test_fraction,
            test_fraction=test_fraction,
            gap_test_end_fraction=gap_test_end_fraction,
            stratified_bool=False,
            y=None,
            stratify_targets_to_ignore=None,
            min_test_size=min_test_size,
        )

        train_fraction_list = validated_dict["train_fraction_list"]
        gap_train_validation_fraction_list = validated_dict[
            "gap_train_validation_fraction_list"
        ]
        validation_fraction_list = validated_dict["validation_fraction_list"]
        gap_validation_test_fraction_list = validated_dict[
            "gap_validation_test_fraction_list"
        ]
        test_fraction_list = validated_dict["test_fraction_list"]
        gap_test_end_fraction_list = validated_dict["gap_test_end_fraction_list"]
        groups_array = validated_dict["groups_array"]
        index_array = validated_dict["index_array"]
        min_test_size_list = validated_dict["min_test_size_list"]

        # iterate over number of folds
        for ii in range(len(train_fraction_list)):
            # fractions for this split
            split_train_fraction = train_fraction_list[ii]
            split_gap_train_val_fraction = gap_train_validation_fraction_list[ii]
            split_validation_fraction = validation_fraction_list[ii]
            split_gap_val_test_fraction = gap_validation_test_fraction_list[ii]
            split_test_fraction = test_fraction_list[ii]
            split_gap_test_end_fraction = gap_test_end_fraction_list[ii]
            split_min_test_size_list = min_test_size_list[ii]

            sorted_groups_indices_dict = make_sorted_groups_indices(
                groups_array=groups_array,
                index_array=index_array,
                targets_array=None,
                target=None,
            )
            sorted_groups_with_nones = sorted_groups_indices_dict[
                "sorted_groups_with_nones"
            ]
            sorted_indices_without_nones = sorted_groups_indices_dict[
                "sorted_indices_without_nones"
            ]
            sorted_indices_with_nones = sorted_groups_indices_dict[
                "sorted_indices_with_nones"
            ]

            begin_end_dict = find_begin_end_indices(
                sorted_indices_without_nones=sorted_indices_without_nones,
                sorted_indices_with_nones=sorted_indices_with_nones,
                split_train_fraction=split_train_fraction,
                split_gap_train_val_fraction=split_gap_train_val_fraction,
                split_validation_fraction=split_validation_fraction,
                split_gap_val_test_fraction=split_gap_val_test_fraction,
                split_test_fraction=split_test_fraction,
                split_gap_test_end_fraction=split_gap_test_end_fraction,
                undefined_groups_in_train=undefined_groups_in_train,
            )
            train_end = begin_end_dict["train_end"]
            validation_begin = begin_end_dict["validation_begin"]
            validation_end = begin_end_dict["validation_end"]
            test_begin = begin_end_dict["test_begin"]
            test_end = begin_end_dict["test_end"]

            # set required min test size (is specified and test_size is not enough)
            if split_min_test_size_list is not None:
                begin_end_dict = force_min_test_size(
                    min_test_size=split_min_test_size_list,
                    train_end=train_end,
                    validation_begin=validation_begin,
                    validation_end=validation_end,
                    test_begin=test_begin,
                    test_end=test_end,
                )
                train_end = begin_end_dict["train_end"]
                validation_begin = begin_end_dict["validation_begin"]
                validation_end = begin_end_dict["validation_end"]
                test_begin = begin_end_dict["test_begin"]
                test_end = begin_end_dict["test_end"]

            # if same group in same set
            if same_group_same_set:
                # change sets to the nearest whole group
                train_end = force_same_set_same_group(
                    sorted_groups_with_nones,
                    train_end,
                )
                validation_begin = force_same_set_same_group(
                    sorted_groups_with_nones,
                    validation_begin,
                )
                validation_end = force_same_set_same_group(
                    sorted_groups_with_nones,
                    validation_end,
                )
                test_begin = force_same_set_same_group(
                    sorted_groups_with_nones,
                    test_begin,
                )
                test_end = force_same_set_same_group(sorted_groups_with_nones, test_end)

            train_indices = sorted_indices_with_nones[:train_end].tolist()
            validation_indices = sorted_indices_with_nones[
                validation_begin:validation_end
            ].tolist()
            test_indices = sorted_indices_with_nones[test_begin:test_end].tolist()

            yield train_indices, validation_indices, test_indices



import numpy as np
import pytest
from numpy.random import default_rng
from scipy.stats import norm

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.uncertainty.calibration_gap import (
    CalibrationGap,
)

rng = default_rng()


@pytest.fixture()
def fixture_metric():
    return load_metric("calibration_gap")


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert "calibration_gap" in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, CalibrationGap)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_standard_deviations(fixture_metric):
    predictions = [0, 0, 0, 1, 1, 1]
    references = [5, 5, 5, 10, 10, 10]
    standard_deviations = [1, 1, 1, 1, 1, 1]

    actual = fixture_metric.compute(
        predictions=predictions,
        references=references,
        standard_deviations=standard_deviations,
    )["calibration_gap"]
    expected = 0.5
    assert np.isclose(actual, expected, atol=0.01)


def test_calibration_gap_calibrated(fixture_metric):
    references = np.concatenate(
        [
            rng.normal(loc=0, scale=1, size=(10000)),
            rng.normal(loc=3, scale=2, size=(10000)),
        ],
    )
    predictions = np.concatenate([np.zeros(10000), 3 * np.ones(10000)])
    standard_deviations = np.concatenate([np.ones(10000), 2 * np.ones(10000)])

    actual = fixture_metric.compute(
        predictions=predictions,
        references=references,
        standard_deviations=standard_deviations,
    )["calibration_gap"]
    expected = 0.0
    assert np.isclose(actual, expected, atol=0.01)


def test_calibration_gap_single_prediction(fixture_metric):
    references = [
        0,
    ]
    predictions = [
        norm.ppf(0.7, loc=0, scale=1),
    ]  # 0.524..., 70% of samples from N(0, 1) are below this.
    standard_deviations = [1.0]

    actual = fixture_metric.compute(
        predictions=predictions,
        references=references,
        standard_deviations=standard_deviations,
        num_thresholds=1001,
    )["calibration_gap"]

    # the calibration curve is 0 <= 0.3 and 1 > 0.3.
    # the area is then the area of two right angle triangles
    # with base and height 0.3 and 0.7 summed.
    expected = 0.5 * (0.7**2 + 0.3**2)
    assert np.isclose(actual, expected, atol=0.01)



from molflux.core.featurisation.featurisation import (
    featurise_dataset,
    replay_dataset_featurisation,
)
from molflux.core.featurisation.metadata import (
    fetch_model_featurisation_metadata,
    load_featurisation_metadata,
)
from molflux.core.models import (
    get_inputs,
    get_references,
    inference,
    load_model,
    predict,
    save_model,
)
from molflux.core.scoring import (
    compute_scores,
    invert_scores_hierarchy,
    merge_scores,
    score_model,
)



from dataclasses import dataclass
from typing import Any, ClassVar, Literal

import h5py

import datasets
from molflux.datasets.typing import ExamplesGenerator

_BASE_URL = "https://zenodo.org/records/8222043/files/SPICE-1.1.4.hdf5?download=1"

_HOMEPAGE = "https://github.com/openmm/spice-dataset"

_DESCRIPTION = """
    SPICE dataset, a new quantum chemistry dataset for training potentials relevant to simulating drug-like small
    molecules interacting with proteins. It contains over 1.1 million conformations for a diverse set of small
    molecules, dimers, dipeptides, and solvated amino acids. It includes 15 molfluxs, charged and uncharged molecules,
    and a wide range of covalent and non-covalent interactions. It provides both forces and energies
    calculated at the ωB97M-D3(BJ)/def2-TZVPPD level of theory.

    WARNING: The molecules here are point clouds. The OpenEye mols do not have any bonds. The charges are added to the
    mols from the smiles provided.
    """


@dataclass
class SPICEConfig(datasets.BuilderConfig):
    backend: Literal["openeye", "rdkit"] = "rdkit"


class SPICE(datasets.GeneratorBasedBuilder):
    BUILDER_CONFIG_CLASS = SPICEConfig
    config: SPICEConfig

    BUILDER_CONFIGS: ClassVar[list[datasets.BuilderConfig]] = [
        SPICEConfig(
            name="openeye",
            backend="openeye",
        ),
        SPICEConfig(
            name="rdkit",
            backend="rdkit",
        ),
    ]

    def _info(self) -> datasets.DatasetInfo:
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(
                {
                    "mol_bytes": datasets.Value("binary"),
                    "dft_total_energy": datasets.Value("float64"),
                    "dft_total_gradient": datasets.Sequence(
                        feature=datasets.Sequence(
                            feature=datasets.Value(dtype="float64", id=None),
                            length=-1,
                            id=None,
                        ),
                        length=-1,
                        id=None,
                    ),
                    "formation_energy": datasets.Value("float64"),
                    "smiles": datasets.Value("string"),
                    "subset": datasets.Value("string"),
                    "mbis_charges": datasets.Sequence(
                        feature=datasets.Sequence(
                            feature=datasets.Value(dtype="float64", id=None),
                            length=-1,
                            id=None,
                        ),
                        length=-1,
                        id=None,
                    ),
                    "mbis_dipoles": datasets.Sequence(
                        feature=datasets.Sequence(
                            feature=datasets.Value(dtype="float64", id=None),
                            length=-1,
                            id=None,
                        ),
                        length=-1,
                        id=None,
                    ),
                    "mbis_quadrupoles": datasets.Sequence(
                        feature=datasets.Sequence(
                            feature=datasets.Sequence(
                                feature=datasets.Value(dtype="float64", id=None),
                                length=-1,
                                id=None,
                            ),
                            length=-1,
                            id=None,
                        ),
                        length=-1,
                        id=None,
                    ),
                    "mbis_octupoles": datasets.Sequence(
                        feature=datasets.Sequence(
                            feature=datasets.Sequence(
                                feature=datasets.Sequence(
                                    feature=datasets.Value(dtype="float64", id=None),
                                    length=-1,
                                    id=None,
                                ),
                                length=-1,
                                id=None,
                            ),
                            length=-1,
                            id=None,
                        ),
                        length=-1,
                        id=None,
                    ),
                    "scf_dipole": datasets.Sequence(
                        feature=datasets.Value(dtype="float64", id=None),
                        length=-1,
                        id=None,
                    ),
                    "scf_quadrupole": datasets.Sequence(
                        feature=datasets.Sequence(
                            feature=datasets.Value(dtype="float64", id=None),
                            length=-1,
                            id=None,
                        ),
                        length=-1,
                        id=None,
                    ),
                    "mayer_indices": datasets.Sequence(
                        feature=datasets.Sequence(
                            feature=datasets.Value(dtype="int64", id=None),
                            length=-1,
                            id=None,
                        ),
                        length=-1,
                        id=None,
                    ),
                    "wiberg_lowdin_indices": datasets.Sequence(
                        feature=datasets.Sequence(
                            feature=datasets.Value(dtype="int64", id=None),
                            length=-1,
                            id=None,
                        ),
                        length=-1,
                        id=None,
                    ),
                },
            ),
            homepage=_HOMEPAGE,
        )

    def _split_generators(
        self,
        dl_manager: datasets.DownloadManager,
    ) -> list[datasets.SplitGenerator]:
        archive_path = dl_manager.download_and_extract(_BASE_URL)

        return [
            datasets.SplitGenerator(
                name="train",
                gen_kwargs={
                    "h5_file_path": archive_path,
                },
            ),
        ]

    def _generate_examples(self, **kwargs: Any) -> ExamplesGenerator:
        if self.config.backend == "openeye":
            return self._generate_examples_openeye(**kwargs)
        elif self.config.backend == "rdkit":
            return self._generate_examples_rdkit(**kwargs)
        else:
            raise KeyError("Unknown backend")

    def _generate_examples_openeye(
        self,
        h5_file_path: str | None = None,
        **kwargs: Any,
    ) -> ExamplesGenerator:
        try:
            from openeye.oechem import OEMol, OESmilesToMol, OEWriteMolToBytes
        except ImportError as e:
            from molflux.datasets.exceptions import ExtrasDependencyImportError

            raise ExtrasDependencyImportError("openeye-toolkits", e) from e

        # open h5 file
        h5_file = h5py.File(h5_file_path, "r+")

        # get columns to loop over
        columns = list(self._info().features.keys())
        columns.remove("mol_bytes")
        columns.append("conformations")

        # reset index
        index = 0

        # loop over chemical formulae (each have multiple conformers)
        for chemical_formula_id in h5_file:
            # the dict of data for this chemical formula
            chemical_formula_dict = {
                key: value[:].tolist()
                for key, value in dict(h5_file[chemical_formula_id]).items()
            }

            # get smiles and subsets for this id
            chemical_formula_id_smiles = chemical_formula_dict["smiles"][0].decode(
                "utf-8",
            )
            chemical_formula_id_subset = chemical_formula_dict["subset"][0].decode(
                "utf-8",
            )
            del chemical_formula_dict["smiles"]
            del chemical_formula_dict["subset"]

            # get charges dictionary from smiles to add to mol
            smiles_mol = OEMol()
            OESmilesToMol(smiles_mol, chemical_formula_id_smiles)
            charge_dict = {}
            for atom in smiles_mol.GetAtoms():
                atom_map_idx = atom.GetMapIdx()
                if atom_map_idx != 0:
                    charge_dict[atom_map_idx - 1] = atom.GetFormalCharge()
                else:
                    raise RuntimeError(
                        "Cannot get atom map idx while extracting charges from smiles.",
                    )

            # atomic numbers list for the chemical formula
            atomic_nums = chemical_formula_dict["atomic_numbers"]
            del chemical_formula_dict["atomic_numbers"]

            # make all the molecules into mol bytes (only atomic numbers and coords available)
            mol_list = []
            for coords in chemical_formula_dict["conformations"]:
                mol = OEMol()
                for idx, atom_num in enumerate(atomic_nums):
                    atom = mol.NewAtom(atom_num)
                    atom.SetFormalCharge(charge_dict[idx])

                flat_coords = [xs for x in coords for xs in x]
                mol.SetCoords(flat_coords)

                mol_bytes = OEWriteMolToBytes(".oeb", mol)
                mol_list.append(mol_bytes)

            # delete coordinates from the dict and add mols
            del chemical_formula_dict["conformations"]
            chemical_formula_dict["mol_bytes"] = mol_list

            # find list of example dicts (swap dict of lists to list of dicts)
            chemical_formula_examples_list = [
                dict(zip(chemical_formula_dict, t, strict=False))
                for t in zip(*chemical_formula_dict.values(), strict=False)
            ]

            # yield the examples for the conformers
            for example_dict in chemical_formula_examples_list:
                index += 1
                example_dict["smiles"] = chemical_formula_id_smiles
                example_dict["subset"] = chemical_formula_id_subset

                # not all subsets have all the columns, fill in with None
                for k in self.info.features:
                    if k not in example_dict:
                        example_dict[k] = None
                yield index, example_dict

    def _generate_examples_rdkit(
        self,
        h5_file_path: str | None = None,
        **kwargs: Any,
    ) -> ExamplesGenerator:
        try:
            from rdkit.Chem import Atom, Conformer, MolFromSmiles, RWMol
            from rdkit.Geometry import Point3D
        except ImportError as e:
            from molflux.datasets.exceptions import ExtrasDependencyImportError

            raise ExtrasDependencyImportError("rdkit", e) from e

        # open h5 file
        h5_file = h5py.File(h5_file_path, "r+")

        # get columns to loop over
        columns = list(self._info().features.keys())
        columns.remove("mol_bytes")
        columns.append("conformations")

        # reset index
        index = 0

        # loop over chemical formulae (each have multiple conformers)
        for chemical_formula_id in h5_file:
            # the dict of data for this chemical formula
            chemical_formula_dict = {
                key: value[:].tolist()
                for key, value in dict(h5_file[chemical_formula_id]).items()
            }

            # get smiles and subsets for this id
            chemical_formula_id_smiles = chemical_formula_dict["smiles"][0].decode(
                "utf-8",
            )
            chemical_formula_id_subset = chemical_formula_dict["subset"][0].decode(
                "utf-8",
            )
            del chemical_formula_dict["smiles"]
            del chemical_formula_dict["subset"]

            # get charges dictionary from smiles to add to mol
            smiles_mol = MolFromSmiles(chemical_formula_id_smiles, sanitize=False)
            charge_dict = {}
            for atom in smiles_mol.GetAtoms():
                atom_map_idx = atom.GetAtomMapNum()
                if atom_map_idx != 0:
                    charge_dict[atom_map_idx - 1] = atom.GetFormalCharge()
                else:
                    raise RuntimeError(
                        "Cannot get atom map idx while extracting charges from smiles.",
                    )

            # atomic numbers list for the chemical formula
            atomic_nums = chemical_formula_dict["atomic_numbers"]
            del chemical_formula_dict["atomic_numbers"]

            # make all the molecules into mol bytes (only atomic numbers and coords available)
            mol_list = []
            for coords in chemical_formula_dict["conformations"]:
                mol = RWMol()
                for idx, atomic_num in enumerate(atomic_nums):
                    atom = Atom(atomic_num)
                    atom.SetFormalCharge(charge_dict[idx])
                    mol.AddAtom(atom)

                conf = Conformer()
                for i in range(mol.GetNumAtoms()):
                    conf.SetAtomPosition(i, Point3D(*coords[i]))

                mol.AddConformer(conf)
                mol_bytes = mol.ToBinary()
                mol_list.append(mol_bytes)

            # delete coordinates from the dict and add mols
            del chemical_formula_dict["conformations"]
            chemical_formula_dict["mol_bytes"] = mol_list

            # find list of example dicts (swap dict of lists to list of dicts)
            chemical_formula_examples_list = [
                dict(zip(chemical_formula_dict, t, strict=False))
                for t in zip(*chemical_formula_dict.values(), strict=False)
            ]

            # yield the examples for the conformers
            for example_dict in chemical_formula_examples_list:
                index += 1
                example_dict["smiles"] = chemical_formula_id_smiles
                example_dict["subset"] = chemical_formula_id_subset

                # not all subsets have all the columns, fill in with None
                for k in self.info.features:
                    if k not in example_dict:
                        example_dict[k] = None
                yield index, example_dict



from typing import Any

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.models.lightning.mlp_regressor.mlp_config import MLPConfig
from molflux.modelzoo.models.lightning.mlp_regressor.mlp_datamodule import (
    MLPDataModule,
)
from molflux.modelzoo.models.lightning.mlp_regressor.mlp_module import MLPModule
from molflux.modelzoo.models.lightning.model import LightningModelBase


class LightningMLPRegressor(LightningModelBase[MLPConfig]):
    """Model class for MLP model"""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description="MLP implementation via lightning",
            config_description="",
        )

    @property
    def _datamodule_builder(self) -> type[MLPDataModule]:
        return MLPDataModule

    @property
    def _config_builder(self) -> type[MLPConfig]:
        return MLPConfig

    def _instantiate_module(self) -> MLPModule:
        return MLPModule(model_config=self.model_config)



"""Median absolute error regression loss."""

import logging
from typing import Any

import evaluate
from sklearn.metrics import median_absolute_error

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
The median_absolute_error is particularly interesting because it is robust to
outliers. The loss is calculated by taking the median of all absolute
differences between the target and the prediction.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated target values.
    references: Ground truth (correct) target values.
    sample_weight (optional): Weighting of each sample.
    multioutput (optional): Defines aggregating of multiple output scores.
        Array-like value defines weights used to average errors. Alternatively,
        one of {'raw_values', 'uniform_average'}. Defaults
        to 'uniform_average'.
        'raw_values' :
            Returns a full set of errors in case of multioutput input.
        'uniform_average' :
            Errors of all outputs are averaged with uniform weight.

Returns:
    median_absolute_error: If multioutput is `raw_values`, then mean absolute
    error is returned for each output separately. If multioutput is
    `uniform_average` or an ndarray of weights, then the weighted average of
    all output errors is returned.

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("median_absolute_error")
    >>> predictions = [2.5, 0.0, 2, 8]
    >>> references = [3, -0.5, 2, 7]
    >>> metric.compute(predictions=predictions, references=references)
    {'median_absolute_error': 0.5}
    >>> metric = load_metric("median_absolute_error", config_name="multioutput")
    >>> predictions = [[0, 2], [-1, 2], [8, -5]]
    >>> references = [[0.5, 1], [-1, 1], [7, -6]]
    >>> metric.compute(predictions=predictions, references=references)
    {'median_absolute_error': 0.75}
    >>> metric.compute(predictions=predictions, references=references, multioutput='raw_values')
    {'median_absolute_error': [0.5, 1.0]}
    >>> metric.compute(predictions=predictions, references=references, multioutput=[0.3, 0.7])
    {'median_absolute_error': 0.85}

"""

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class MedianAbsoluteError(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("float")),
                    "references": datasets.Sequence(datasets.Value("float")),
                }
                if self.config_name == "multioutput"
                else {
                    "predictions": datasets.Value("float"),
                    "references": datasets.Value("float"),
                },
            ),
            reference_urls=[
                "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.r2_score.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        sample_weight: list[float] | None = None,
        multioutput: str = "uniform_average",
        **kwargs: Any,
    ) -> MetricResult:
        score = median_absolute_error(
            y_true=references,
            y_pred=predictions,
            sample_weight=sample_weight,
            multioutput=multioutput,
        ).tolist()
        return {self.tag: score}



import pandas as pd

import datasets
from molflux.modelzoo.interchange import hf_dataset_from_dataframe


def test_is_noop_on_huggingface_datasets():
    """That the interchange function is pass-through for huggingface Datasets."""
    dataset = datasets.Dataset.from_dict({"A": [1, 2, 3], "B": [4, 5, 6]})
    out = hf_dataset_from_dataframe(dataset)
    assert out == dataset


def test_can_interchange_pandas_df_with_nested_fields():
    """That can interchange a pandas DataFrame with array-like features."""
    dataframe = pd.DataFrame({"A": [1, 2], "B": [[1, 2, 3], [4, 5, 6]]})
    out = hf_dataset_from_dataframe(dataframe)
    assert isinstance(out, datasets.Dataset)



import pytest

from molflux.features import load_representation


@pytest.fixture(scope="module")
def fixture_mock_representation():
    return load_representation(name="character_count")


def test_on_array_returns_correct_dimensions(fixture_mock_representation):
    """That array featurisation returns an output of the same shape."""
    representation = fixture_mock_representation
    data = ["cccc", "cc"]
    representation_result = representation.featurise(data)
    featurised_data = representation_result.get(representation.tag)
    assert len(featurised_data) == len(data)


def test_on_single_sample_returns_correct_dimensions(fixture_mock_representation):
    """That featurisation of a single sample returns a single result.

    This is to avoid bugs where single sample iterable is iterated over.
    """
    representation = fixture_mock_representation
    data = "cccc"
    representation_result = representation.featurise(data)
    featurised_data = representation_result.get(representation.tag)
    assert len(featurised_data) == 1


def test_can_featurise_empty_data(fixture_mock_representation):
    """That can featurise an empty array without problems."""
    representation = fixture_mock_representation
    assert representation.featurise([])


def test_invalid_kwarg_raises_value_error(fixture_mock_representation):
    """That attempting featurisation with invalid keyword arguments raises."""
    representation = fixture_mock_representation
    data = ["cccc", "cc"]
    with pytest.raises(ValueError, match=r"Unknown featurisation parameter\(s\)"):
        representation.featurise(data, invalid_kwarg=True)



"""
Tests for API features advertised as part of top level package namespace.
"""

import molflux.splits


def test_exports_list_splitting_strategies():
    """That the package exposes the list_splitting_strategies function."""
    assert hasattr(molflux.splits, "list_splitting_strategies")


def test_exports_k_fold_split():
    """That the package exposes the k_fold_split function."""
    assert hasattr(molflux.splits, "k_fold_split")


def test_exports_train_test_split():
    """That the package exposes the train_test_split function."""
    assert hasattr(molflux.splits, "train_test_split")


def test_exports_train_validation_test_split():
    """That the package exposes the k_fold_split function."""
    assert hasattr(molflux.splits, "train_validation_test_split")


def test_exports_load_from_dict():
    """That the package exposes the load_from_dict function."""
    assert hasattr(molflux.splits, "load_from_dict")


def test_exports_load_from_dicts():
    """That the package exposes the load_from_dicts function."""
    assert hasattr(molflux.splits, "load_from_dicts")


def test_exports_load_from_yaml():
    """That the package exposes the load_from_yaml function."""
    assert hasattr(molflux.splits, "load_from_yaml")


def test_exports_load_splitting_strategy():
    """That the package exposes the load_splitting_strategy function."""
    assert hasattr(molflux.splits, "load_splitting_strategy")


def test_exports_splitting_strategy():
    """That the package exposes the SplittingStrategy protocol."""
    assert hasattr(molflux.splits, "SplittingStrategy")



from typing import Any

try:
    from rdkit.Chem.MACCSkeys import GenMACCSKeys

    from molflux.features.representations.rdkit._utils import rdkit_mol_from_smiles
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import to_smiles
from molflux.features.typing import Fingerprint, MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
MACCS fingerprint. [rdkit]

MACCS keys are 166 bit structural key descriptors in which each bit is
associated with a SMARTS pattern.
"""


class MACCSRdkit(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        **kwargs: Any,
    ) -> dict[str, list[Fingerprint]]:
        """Generates MACCS fingerprints for each input molecule.

        Args:
            samples: The molecules to be fingerprinted.

        Returns:
            MACCS fingerprints, as lists of bits.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('maccs_rdkit')
            >>> samples = ['c1ccccc1']
            >>> representation.featurise(samples)
            {'maccs_rdkit': [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0]]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        maccs_fp_list: list[list] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                smile = to_smiles(sample)
                mol = rdkit_mol_from_smiles(smile)
                rd_fp = GenMACCSKeys(mol)

                if rd_fp.GetNumOnBits() == 0:
                    bit_vector = [0] * rd_fp.GetNumBits()
                else:
                    bit_vector = rd_fp.ToList()

                maccs_fp_list.append(bit_vector[1:])

        return {self.tag: maccs_fp_list}



import numpy as np
import pytest
import torch
from torch._dynamo.testing import CompileCounter

from molflux.modelzoo import load_from_store, save_to_store
from molflux.modelzoo.catalogue import list_models
from molflux.modelzoo.models.lightning.config import CompileConfig
from molflux.modelzoo.models.lightning.mlp_regressor.mlp_model import (
    LightningMLPRegressor,
)
from molflux.modelzoo.protocols import Estimator

model_name = "lightning_mlp_regressor"


def test_in_catalogue():
    """That the model is registered in the catalogue."""
    catalogue = list_models()
    all_names = [name for names in catalogue.values() for name in names]
    assert model_name in all_names


def test_default_model_tag_matches_entrypoint_name(fixture_model):
    """That the default model tag matches the catalogue entrypoint name.

    This is not strictly required, but ensures a more intuitive user experience.
    """
    model = fixture_model
    assert model.tag == model_name


def test_is_mapped_to_correct_class(fixture_model):
    """That the model name is mapped to the appropriate class."""
    model = fixture_model
    assert isinstance(model, LightningMLPRegressor)


def test_implements_protocol(fixture_model):
    """That the model implements the public Estimator protocol."""
    model = fixture_model
    assert isinstance(model, Estimator)


def test_train_predict_model(
    fixture_model,
    tmp_path,
    train_dataset,
    validation_dataset,
    predict_dataset,
):
    """Test that a model can run the train and predict functions"""

    model = fixture_model

    model.train(
        train_data=train_dataset,
        validation_data=validation_dataset,
        trainer_config={
            "accelerator": "cpu",
            "max_epochs": 5,
            "default_root_dir": str(tmp_path),
        },
        datamodule_config={
            "train": {"batch_size": 2},
            "validation": {"batch_size": 2},
        },
    )
    predictions = model.predict(
        predict_dataset,
        datamodule_config={"predict": {"batch_size": 2}},
    )
    assert predictions
    assert len(predictions) == len(model.y_features)
    for task_predictions in predictions.values():
        assert len(task_predictions) == len(predict_dataset)


def test_train_multi_data(
    fixture_model,
    tmp_path,
    train_dataset,
    train_dataset_b,
    validation_dataset,
):
    """Test that a model can train with multiple datasets"""

    model = fixture_model

    # NOTE train_dataset has 5 examples, train_dataset_b has 8
    model.train(
        train_data={"a": train_dataset, "b": train_dataset_b},
        validation_data={"a": validation_dataset, "c": validation_dataset},
        trainer_config={
            "accelerator": "cpu",
            "max_epochs": 5,
            "default_root_dir": str(tmp_path),
        },
        datamodule_config={
            "train": {"batch_size": 2},
            "validation": {"batch_size": 2},
        },
    )


def test_saving_loading(
    tmp_path,
    fixture_model,
    train_dataset,
    validation_dataset,
    predict_dataset,
):
    """Test that a model can successfully be saved and loaded"""

    fixture_model.train(
        train_data=train_dataset,
        validation_data=validation_dataset,
        trainer_config={
            "accelerator": "cpu",
            "max_epochs": 5,
            "default_root_dir": str(tmp_path),
        },
        datamodule_config={
            "train": {"batch_size": 2},
            "validation": {"batch_size": 2},
        },
    )

    save_to_store(tmp_path, fixture_model)
    loaded_model = load_from_store(tmp_path)

    assert loaded_model.predict(
        predict_dataset,
        datamodule_config={
            "predict": {"batch_size": 2},
        },
    )


@pytest.mark.parametrize("compile", [True, False])
def test_does_compile(
    tmp_path,
    fixture_model,
    train_dataset,
    validation_dataset,
    compile,
    compile_error_context,
):
    """Test that the model does compile during training when set to"""
    # Python 3.11 does not support compilation as of PyTorch 2.0.1, but 2.1 will

    assert not fixture_model.is_compiled

    if compile:
        cnt = CompileCounter()
        fixture_model.model_config.compile = CompileConfig("default")
        fixture_model.model_config.compile.backend = cnt  # pyright: ignore

        assert cnt.frame_count == 0

    with compile_error_context:
        fixture_model.train(
            train_data=train_dataset,
            validation_data=validation_dataset,
            trainer_config={
                "accelerator": "cpu",
                "max_epochs": 5,
                "default_root_dir": str(tmp_path),
            },
            datamodule_config={
                "train": {"batch_size": 2},
                "validation": {"batch_size": 2},
            },
        )

        if compile:
            assert fixture_model.is_compiled
            assert cnt.frame_count > 0  # pyright: ignore


def test_train_model_overfit(
    fixture_model,
    tmp_path,
    train_dataset,
):
    """Test that a model can overfit to a small batch"""

    torch.manual_seed(0)

    model = fixture_model

    model.train(
        train_data=train_dataset,
        validation_data=None,
        trainer_config={
            "accelerator": "cpu",
            "max_epochs": 400,
            "default_root_dir": str(tmp_path),
        },
        datamodule_config={
            "train": {"batch_size": 2},
            "validation": {"batch_size": 2},
        },
        optimizer_config={"name": "SGD", "config": {"lr": 1e-3}},
        scheduler_config={
            "name": "OneCycleLR",
            "config": {
                "max_lr": 1e-3,
                "total_steps": "num_steps",
                "pct_start": 0.1,
                "anneal_strategy": "linear",
            },
        },
    )
    predictions = model.predict(
        train_dataset,
        datamodule_config={"predict": {"batch_size": 2}},
    )

    np.testing.assert_allclose(
        predictions["lightning_mlp_regressor::y_col"],
        train_dataset["y_col"],
        rtol=5e-2,
    )



"""Gaussain NLL (Gaussian Negative Log Likelihood) regression score function."""

import logging
from typing import Any

import evaluate
import numpy as np

import datasets
from molflux.metrics.bases import UncertaintyMetric
from molflux.metrics.typing import ArrayLike, MetricResult
from molflux.metrics.uncertainty.utils import _estimate_standard_deviation

logger = logging.getLogger(__name__)


_DESCRIPTION = """
Gaussian Negative Log Likelihood function.

The predictions are treated as samples from Gaussian distributions with a given
expectation and variance. The probability density function of the Gaussian is computed
at the reference point. A minus log is applied, and values for multiple data points
are added together.

"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: array of mean predictions.
    references: Ground truth (correct) target values.
    prediction_intervals: Prediction intervals of lower and upper bounds.
    confidence (optional): float indicating what proportion of data lies between
        lower and upper bounds. Defaults to 0.9.
    eps (optional): Minimum value for standard deviations to avoid numeric errors.
        Defaults to 1e-6.

Returns:
    gaussian_nll: The Gaussian NLL value.

Examples:
    >>> from molflux.metrics import load_metric
    >>> m = load_metric("gaussian_nll")
    >>> ref = [0, 1, 2, 3, 4]
    >>> pred = [0.5, 0.8, 1.2, 2.4, 4.5]
    >>> lower_bound = [0.1, 0.1, 0.2, 2, 3]
    >>> upper_bound = [1, 1.5, 1.8, 2.8, 5.7]
    >>> prediction_intervals = list(zip(lower_bound, upper_bound))
    >>> m.compute(predictions=pred, references=ref, prediction_intervals=prediction_intervals)
    {'gaussian_nll': 1.2949031482515838}
"""

_CITATION = """\
None
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class GaussianNLL(UncertaintyMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Value("float64"),
                    "references": datasets.Value("float64"),
                },
            ),
            reference_urls=["https://en.wikipedia.org/wiki/Likelihood_function"],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        standard_deviations: ArrayLike | None = None,
        prediction_intervals: ArrayLike | None = None,
        confidence: float = 0.9,
        eps: float = 1e-6,
        **kwargs: Any,
    ) -> MetricResult:
        mu = np.array(predictions)

        if standard_deviations is None:
            if prediction_intervals is None:
                raise ValueError(
                    "Please provide prediction intervals in the form of lower and upper bounds.",
                )
            lower_bound, upper_bound = zip(*prediction_intervals, strict=False)
            lower_bound = np.array(lower_bound)  # type: ignore[assignment]
            upper_bound = np.array(upper_bound)  # type: ignore[assignment]
            if (upper_bound < lower_bound).any():  # type: ignore[attr-defined]
                raise ValueError(
                    "Please ensure upper bound is greater than lower bound.",
                )

            sigma = _estimate_standard_deviation(lower_bound, upper_bound, confidence)
        else:
            sigma = np.array(standard_deviations)

        sigma = np.clip(sigma, a_min=eps, a_max=None)

        gaussian_log_likelihood = (
            1
            / 2
            * np.mean(
                np.log(2 * np.pi * (sigma**2)) + ((references - mu) ** 2) / (sigma**2),
            )
        )

        return {self.tag: gaussian_log_likelihood}



"""Explained variance regression score function."""

import logging
from typing import Any

import evaluate
from sklearn.metrics import explained_variance_score

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
The explained variance computes the explained variance regression score.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated target values.
    references: Ground truth (correct) target values.
    sample_weight (optional): Weighting of each sample.
    multioutput (optional): Defines aggregating of multiple output scores.
        Array-like value defines weights used to average errors. Alternatively,
        one of {'raw_values', 'uniform_average', 'variance_weighted'}. Defaults
        to 'uniform_average'.
        'raw_values' :
            Returns a full set of errors in case of multioutput input.
        'uniform_average' :
            Errors of all outputs are averaged with uniform weight.
        'variance_weighted' :
            Scores of all outputs are averaged, weighted by the variances
            of each individual output.

Returns:
    explained_variance: The explained variance or ndarray if 'multioutput' is
    'raw_values'.

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("explained_variance")
    >>> predictions = [2.5, 0.0, 2, 8]
    >>> references = [3, -0.5, 2, 7]
    >>> metric.compute(predictions=predictions, references=references)
    {'explained_variance': 0.957...}
    >>> metric = load_metric("explained_variance", config_name="multioutput")
    >>> predictions = [[0, 2], [-1, 2], [8, -5]]
    >>> references = [[0.5, 1], [-1, 1], [7, -6]]
    >>> metric.compute(predictions=predictions, references=references, multioutput='uniform_average')
    {'explained_variance': 0.983...}
"""

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class ExplainedVariance(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("float")),
                    "references": datasets.Sequence(datasets.Value("float")),
                }
                if self.config_name == "multioutput"
                else {
                    "predictions": datasets.Value("float"),
                    "references": datasets.Value("float"),
                },
            ),
            reference_urls=[
                "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.explained_variance_score.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        sample_weight: list[float] | None = None,
        multioutput: str = "uniform_average",
        **kwargs: Any,
    ) -> MetricResult:
        score = explained_variance_score(
            y_true=references,
            y_pred=predictions,
            sample_weight=sample_weight,
            multioutput=multioutput,
        )
        return {self.tag: score}



"""Uncertainty measure: Computes coverage of a prediction interval."""

import logging
from typing import Any

import evaluate
import numpy as np
import scipy

import datasets
from molflux.metrics.bases import UncertaintyMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_CITATION = """\
@article{barber2021predictive,
  title={Predictive inference with the jackknife+},
  author={Barber, Rina Foygel and Candes, Emmanuel J and Ramdas, Aaditya and Tibshirani, Ryan J},
  journal={The Annals of Statistics},
  volume={49},
  number={1},
  pages={486--507},
  year={2021},
  publisher={Institute of Mathematical Statistics}
}
"""

_DESCRIPTION = """\
Computes coverage of a prediction interval.
This is obtained by estimating the fraction of true labels that lie within the prediction intervals.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated targets as returned by a classifier/regressor.
    references: Ground truth (correct) target values.
    prediction_intervals: Prediction intervals of lower and upper bounds.

Returns:
    prediction_interval_coverage: the coverage indicating the fraction of ground truth values that lie in the prediction interval

Examples:
    >>> from molflux.metrics import load_metric
    >>> m = load_metric("prediction_interval_coverage")
    >>> ref = [0, 1, 2, 3, 4]
    >>> pred = [0, 2, 2, 3, 5]
    >>> lower_bound = [0.1, 0.1, 0.2, 2, 3]
    >>> upper_bound = [1, 1.5, 1.8, 2.8, 5.7]
    >>> prediction_intervals = list(zip(lower_bound, upper_bound))
    >>> m.compute(references=ref, predictions=pred, prediction_intervals=prediction_intervals)
    {'prediction_interval_coverage': 0.4}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class PredictionIntervalCoverage(UncertaintyMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Value("float"),
                    "references": datasets.Value("float"),
                },
            ),
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        standard_deviations: ArrayLike | None = None,
        prediction_intervals: ArrayLike | None = None,
        confidence: float = 0.9,
        **kwargs: Any,
    ) -> MetricResult:
        if standard_deviations is not None and prediction_intervals is not None:
            raise ValueError(
                "Both standard deviations and prediction intervals given. Please provide only one.",
            )

        if standard_deviations is not None:
            lower_bound, upper_bound = scipy.stats.norm.interval(
                confidence,
                loc=predictions,
                scale=standard_deviations,
            )
        elif prediction_intervals is not None:
            lower_bound, upper_bound = zip(*prediction_intervals, strict=False)
        else:
            raise ValueError(
                "Either standard_deviations or prediction_intervals must be supplied.",
            )

        lower_bound, upper_bound = np.array(lower_bound), np.array(upper_bound)
        score = np.mean((lower_bound <= references) & (upper_bound >= references))
        return {self.tag: float(score)}



from typing import Any

try:
    from openeye.oemolprop import OEGetLipinskiAcceptorCount
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("openeye", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.openeye._utils import to_digit, to_oemol
from molflux.features.typing import MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
The number of acceptors in a molecule based on the definition from the work of
Lipinski ([Lipinski-1997]).

It is defined to be the number of nitrogens or oxygens.
"""


class NumAcceptors(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        digitise: bool = False,
        start: int = 0,
        stop: int = 10,
        num: int = 10,
        **kwargs: Any,
    ) -> dict[str, list[int]]:
        """Calculates the number of Lipinski acceptors in a molecule.

        Args:
            samples: The sample molecules to featurise.
            digitise: Whether to digitise the results. Defaults to `False`.
            start: If `digitise=True`, the start of the binning range. Defaults to 0.
            stop: If `digitise=True`, the end of the binning range. Defaults to 10.
            num: If `digitise=True`, the number of bins. Defaults to 10.

        Returns:
            The number of acceptors for each input molecule.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('num_acceptors')
            >>> samples = ['c1ccccc1', "NH"]
            >>> representation.featurise(samples=samples)
            {'num_acceptors': [0, 1]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        results = []
        for sample in samples:
            with featurisation_error_harness(sample):
                mol = to_oemol(sample)
                result = OEGetLipinskiAcceptorCount(mol)

                if digitise:
                    result = to_digit(result, start=start, stop=stop, num=num).tolist()

                results.append(result)

        return {self.tag: results}



import logging
from dataclasses import asdict, field
from typing import Any, Literal, TypeVar

from pydantic.v1 import dataclasses, validator

from molflux.modelzoo.model import ModelConfig

try:
    from class_resolver import ClassResolver
    from class_resolver.contrib.torch import lr_scheduler_resolver
    from lightning.pytorch import Trainer
    from lightning.pytorch import callbacks as pl_callbacks
    from lightning.pytorch import loggers as pl_loggers
    from lightning.pytorch import profilers as pl_profilers
    from lightning.pytorch import strategies as pl_strategies
    from torch.optim import Optimizer

    from molflux.modelzoo.models.lightning.trainer.callbacks import callback_resolver
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("lightning", e) from e


logger_resolver = ClassResolver.from_subclasses(
    pl_loggers.Logger,
)

try:
    from dvclive.lightning import DVCLiveLogger  # pyright: ignore

    logger_resolver.register(DVCLiveLogger)

    logging.getLogger("dvc_studio_client.post_live_metrics").setLevel(
        logging.WARNING,
    )

except ImportError:
    pass


profiler_resolver = ClassResolver.from_subclasses(
    pl_profilers.Profiler,
)

strategy_resolver = ClassResolver.from_subclasses(
    pl_strategies.Strategy,
)


class ConfigDict:
    extra = "forbid"
    arbitrary_types_allowed = True
    smart_union = True


def _dict_is_single_logger_config(logger_dict: dict[str, Any]) -> bool:
    return ("name" in logger_dict) and (isinstance(logger_dict["name"], str))


@dataclasses.dataclass(config=ConfigDict)
class TrainerConfig:
    """
    Lightning Trainer config.
    See https://lightning.ai/docs/pytorch/stable/common/trainer.html#trainer-class-api for more info.
    """

    accelerator: str = "auto"
    accumulate_grad_batches: int = 1
    barebones: bool = False
    benchmark: bool | None = None
    callbacks: list[dict[str, Any]] | dict[str, dict[str, Any]] | None = None
    check_val_every_n_epoch: int | None = 1
    default_root_dir: str | None = "training"
    detect_anomaly: bool = False
    deterministic: bool | Literal["warn"] | None = None
    devices: list[int] | str | int = "auto"
    enable_checkpointing: bool = False  # NOTE vanilla PL default is True
    enable_model_summary: bool = True
    enable_progress_bar: bool = True
    fast_dev_run: int | bool = False
    gradient_clip_algorithm: str | None = None
    gradient_clip_val: int | float | None = None
    inference_mode: bool = True
    limit_predict_batches: int | float | None = None
    limit_test_batches: int | float | None = None
    limit_train_batches: int | float | None = None
    limit_val_batches: int | float | None = None
    log_every_n_steps: int = 50
    logger: (
        dict[str, Any] | list[dict[str, Any]] | dict[str, dict[str, Any]] | bool | None
    ) = True
    max_epochs: int | None = 1
    max_steps: int = -1
    max_time: str | dict[str, int] | None = None
    min_epochs: int | None = None
    min_steps: int | None = None
    num_nodes: int = 1
    num_sanity_val_steps: int | None = None
    overfit_batches: int | float = 0.0
    precision: int | str = 32
    profiler: dict[str, Any] | None = None
    reload_dataloaders_every_n_epochs: int = 0
    strategy: str | dict[str, Any] = "auto"
    sync_batchnorm: bool = False
    use_distributed_sampler: bool = True
    val_check_interval: int | float | None = None

    def pass_to_trainer(self) -> dict[str, Any]:
        out = asdict(self)
        out["callbacks"] = self.convert_callbacks(out["callbacks"])
        out["logger"] = self.convert_logger(out["logger"])
        out["profiler"] = self.convert_profiler(out["profiler"])
        out["strategy"] = self.convert_strategy(out["strategy"])

        return out

    def convert_callbacks(
        self,
        callbacks: list[dict[str, Any]] | dict[str, dict[str, Any]] | None,
    ) -> list[pl_callbacks.Callback] | None:
        if callbacks is None:
            return callbacks

        if isinstance(callbacks, dict):
            callbacks = list(callbacks.values())

        return [
            callback_resolver.make(callback["name"], callback.get("config", {}))
            for callback in callbacks
        ]

    def convert_logger(
        self,
        logger: dict[str, Any]
        | list[dict[str, Any]]
        | dict[str, dict[str, Any]]
        | bool
        | None,
    ) -> list[pl_loggers.Logger] | bool | None:
        if logger is None or isinstance(logger, bool):
            return logger

        if isinstance(logger, dict):
            if _dict_is_single_logger_config(logger):
                logger = [logger]
            else:
                logger = list(logger.values())

        return [
            logger_resolver.make(logger_["name"], logger_.get("config", {}))
            for logger_ in logger
        ]

    def convert_profiler(
        self,
        profiler: dict[str, Any] | None,
    ) -> pl_profilers.Profiler | None:
        if profiler is None:
            return profiler

        return profiler_resolver.make(profiler["name"], profiler.get("config", {}))

    def convert_strategy(
        self,
        strategy: dict[str, Any] | str,
    ) -> pl_strategies.Strategy | str:
        if isinstance(strategy, str):
            return strategy

        return strategy_resolver.make(strategy["name"], strategy.get("config", {}))


@dataclasses.dataclass(config=ConfigDict)
class OptimizerConfig:
    name: str = "Adam"
    config: dict[str, Any] = field(default_factory=lambda: {"lr": 1e-4})


@dataclasses.dataclass(config=ConfigDict)
class SchedulerConfig:
    name: str = "CosineAnnealingLR"
    config: dict[str, Any] = field(default_factory=lambda: {"T_max": "num_steps"})
    interval: Literal["epoch", "step"] = "step"
    frequency: int = 1
    monitor: str = "val_loss"
    strict: bool = True

    def prepare_scheduler(
        self,
        optimizer: Optimizer,
        trainer: Trainer,
    ) -> dict[str, Any]:
        # NOTE Enables config-driven decay of learning rate to final step
        if self.config.get("T_max", None) == "num_steps":
            self.config["T_max"] = trainer.estimated_stepping_batches
            self.interval = "step"
        elif self.config.get("T_max", None) == "num_epochs":
            self.config["T_max"] = trainer.max_epochs
            self.interval = "epoch"

        if self.config.get("total_steps", None) == "num_steps":
            self.config["total_steps"] = trainer.estimated_stepping_batches
            self.interval = "step"
        elif self.config.get("total_steps", None) == "num_epochs":
            self.config["total_steps"] = trainer.max_epochs
            self.interval = "epoch"

        return {
            "scheduler": lr_scheduler_resolver.make(
                self.name,
                self.config,
                optimizer=optimizer,
            ),
            "interval": self.interval,
            "frequency": self.frequency,
            "monitor": self.monitor,
            "strict": self.strict,
        }


@dataclasses.dataclass(config=ConfigDict)
class TransferLearningStage:
    freeze_modules: list[str] | None = None
    trainer: dict[str, Any] | None = None
    datamodule: dict[str, Any] | None = None
    optimizer: dict[str, Any] | None = None
    scheduler: dict[str, Any] | None = None

    @validator("trainer")
    def must_fit_into_trainer(
        cls,
        trainer: dict[str, Any] | None,
    ) -> dict[str, Any] | None:
        if trainer is not None:
            TrainerConfig(**trainer)
        return trainer

    @validator("datamodule")
    def must_fit_into_datamodule(
        cls,
        datamodule: dict[str, Any] | None,
    ) -> dict[str, Any] | None:
        if datamodule is not None:
            DataModuleConfig(**datamodule)
        return datamodule

    @validator("optimizer")
    def must_fit_into_optimizer(
        cls,
        optimizer: dict[str, Any] | None,
    ) -> dict[str, Any] | None:
        if optimizer is not None:
            OptimizerConfig(**optimizer)
        return optimizer

    @validator("scheduler")
    def must_fit_into_scheduler(
        cls,
        scheduler: dict[str, Any] | None,
    ) -> dict[str, Any] | None:
        if scheduler is not None:
            SchedulerConfig(**scheduler)
        return scheduler


@dataclasses.dataclass(config=ConfigDict)
class TransferLearningConfigBase:
    stages: list[TransferLearningStage]
    pre_trained_model_path: str | None = None
    repo_url: str | None = None
    rev: str | None = None
    model_path_in_repo: str = "model"
    modules_to_match: dict[str, str] | None = None

    def __post_init__(self) -> None:
        if (self.pre_trained_model_path is not None) ^ (
            (self.repo_url is not None)
            and (self.rev is not None)
            and (self.model_path_in_repo is not None)
        ):
            pass
        else:
            raise ValueError(
                "Must specify either 'pre_trained_model_path' or all of 'repo_url', 'rev', 'model_path_in_repo' (but not both).",
            )

        for i in range(len(self.stages)):
            if isinstance(self.stages[i], dict):
                self.stages[i] = TransferLearningStage(**self.stages[i])  # type: ignore


@dataclasses.dataclass(config=ConfigDict)
class SplitConfig:
    """batch_size: The batch size to use for this data split, optionally as a
    dictionary assigning different batch sizes to different datasets."""

    batch_size: int | dict[str, int] = 1


@dataclasses.dataclass(config=ConfigDict)
class TrainSplitConfig(SplitConfig):
    drop_last: bool = True
    mode: Literal[
        "max_size_cycle",
        "max_size",
        "min_size",
    ] = "max_size_cycle"


@dataclasses.dataclass(config=ConfigDict)
class EvalSplitConfig(SplitConfig):
    mode: Literal[
        "max_size_cycle",
        "max_size",
        "min_size",
        "sequential",
    ] = "max_size"


@dataclasses.dataclass(config=ConfigDict)
class DataModuleConfig:
    train: TrainSplitConfig = field(default_factory=TrainSplitConfig)
    validation: EvalSplitConfig = field(default_factory=EvalSplitConfig)
    test: EvalSplitConfig = field(default_factory=EvalSplitConfig)
    predict: SplitConfig = field(default_factory=SplitConfig)
    num_workers: int | Literal["all"] = 0


@dataclasses.dataclass(config=ConfigDict)
class CompileConfig:
    mode: Literal["default", "reduce-overhead", "max-autotune"]
    dynamic: bool = False
    fullgraph: bool = False
    backend: str = "inductor"
    backend_kwargs: dict[str, Any] = field(default_factory=dict)


@dataclasses.dataclass(config=ConfigDict)
class LightningConfig(ModelConfig):
    datamodule: DataModuleConfig = field(default_factory=DataModuleConfig)
    trainer: TrainerConfig = field(default_factory=TrainerConfig)
    optimizer: OptimizerConfig = field(default_factory=OptimizerConfig)
    scheduler: SchedulerConfig | None = None
    transfer_learning: TransferLearningConfigBase | None = None
    compile: CompileConfig | bool = False
    # NOTE This setting affects a global variable!
    float32_matmul_precision: Literal["highest", "high", "medium"] | None = None

    @validator("compile", pre=True)
    def compile_true_to_compile_config_with_default_mode(
        cls,
        compile: dict[str, Any] | bool,
    ) -> dict[str, Any] | bool:
        if compile is True:
            return {"mode": "default"}
        return compile


LightningConfigT = TypeVar("LightningConfigT", bound=LightningConfig)



from abc import abstractmethod
from collections.abc import Mapping
from typing import Any, Literal, TypeAlias, cast

from lightning.pytorch.utilities.types import _METRIC

from molflux.modelzoo.models.lightning.config import LightningConfig

try:
    import torch
    from class_resolver.contrib.torch import optimizer_resolver
    from lightning.pytorch import LightningModule
    from lightning.pytorch.utilities import grad_norm
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("lightning", e) from e


Loss: TypeAlias = torch.Tensor
TensorsToLog: TypeAlias = dict[str, torch.Tensor]
BatchSize: TypeAlias = int
SingleBatchStepOutput: TypeAlias = tuple[Loss, TensorsToLog, BatchSize]
Split = Literal["train", "val"]


class LightningModuleBase(LightningModule):
    def __init__(
        self,
        model_config: LightningConfig,
    ) -> None:
        super().__init__()

        self.model_config = model_config

    @abstractmethod
    def _training_step_on_single_source_batch(
        self,
        single_source_batch: Any,
        source_name: str | None,
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> SingleBatchStepOutput:
        """Implements a training step on a subbatch from a single dataloader.

        Must return a loss, a (possibly empty) dictionary of tensors, which will be
        automatically logged, and a batch size."""
        ...

    @abstractmethod
    def _validation_step_on_single_source_batch(
        self,
        single_source_batch: Any,
        source_name: str | None,
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> SingleBatchStepOutput:
        """Implements a validation step on a subbatch from a single dataloader.

        Must return a loss, a (possibly empty) dictionary of tensors, which will be
        automatically logged, and a batch size."""
        ...

    def training_step(
        self,
        batch: dict[str | None, Any],
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> torch.Tensor:
        """Implements a single training step.

        Overrides an abstract method in PyTorch Lightning. Implementations
        probably don't need to modify this method; look at
        _training_step_on_single_source_batch instead."""

        # compute losses
        losses: dict[str | None, dict[str, torch.Tensor]] = {}
        batch_sizes: dict[str | None, int] = {}
        for name, single_source_batch in batch.items():
            # can also get {"a": single_source_batch, "b": None} batches
            if single_source_batch is None:
                continue

            (
                loss,
                losses[name],
                batch_sizes[name],
            ) = self._training_step_on_single_source_batch(
                single_source_batch,
                name,
                batch_idx,
                *args,
                **kwargs,
            )
            losses[name]["loss"] = loss

        total_loss = self._process_and_log_losses("train", losses, batch_sizes)
        return total_loss

    def validation_step(
        self,
        batch: dict[str | None, Any],
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> torch.Tensor:
        """Implements a single training step.

        Overrides an abstract method in PyTorch Lightning. Implementations
        probably don't need to modify this method; look at
        _validation_step_on_single_source_batch instead."""

        # compute losses
        losses: dict[str | None, dict[str, torch.Tensor]] = {}
        batch_sizes: dict[str | None, int] = {}
        for name, single_source_batch in batch.items():
            # can also get {"a": single_source_batch, "b": None} batches
            if single_source_batch is None:
                continue

            (
                loss,
                losses[name],
                batch_sizes[name],
            ) = self._validation_step_on_single_source_batch(
                single_source_batch,
                name,
                batch_idx,
                *args,
                **kwargs,
            )
            losses[name]["loss"] = loss

        total_loss = self._process_and_log_losses("val", losses, batch_sizes)
        return total_loss

    def configure_optimizers(self) -> dict[str, Any]:
        """Returns an optimizer and (optional) scheduler.

        Expected by PyTorch Lightning.
        https://lightning.ai/docs/pytorch/stable/api/lightning.pytorch.core.LightningModule.html#lightning.pytorch.core.LightningModule.configure_optimizers
        """

        if self.model_config.optimizer is None:
            raise ValueError("Optimizer must be specified in model config.")

        optimizer_config = self.model_config.optimizer
        optimizer = optimizer_resolver.make(
            optimizer_config.name,
            **optimizer_config.config,
            params=filter(lambda p: p.requires_grad, self.parameters()),
        )
        out: dict[str, Any] = {"optimizer": optimizer}

        if self.model_config.scheduler is not None:
            out["lr_scheduler"] = self.model_config.scheduler.prepare_scheduler(
                optimizer=optimizer,
                trainer=self.trainer,
            )

        return out

    def on_before_optimizer_step(self, optimizer: torch.optim.Optimizer) -> None:
        """Logs gradient magnitudes during training."""
        del optimizer

        grads = cast(dict[str, torch.Tensor], grad_norm(self, norm_type=2))
        # `grads` includes individual layer grads as well as an overall Frobenius norm
        grad_norm_total = grads.pop("grad_2.0_norm_total").item()
        grad_norm_max = max(t.item() for t in grads.values())
        self.log_dict(
            {"grad_norm_total": grad_norm_total, "grad_norm_max": grad_norm_max},
            on_step=True,
            on_epoch=False,
        )

    def log_by_split(
        self,
        split: Split,
        source: str,
        name: str,
        value: _METRIC,
        batch_size: int | None = None,
    ) -> None:
        """Convenience method for logging."""
        if split == "train":
            self.log(
                f"{split}/{source}/{name}",
                value,
                prog_bar=True,
                on_step=True,
                on_epoch=False,
                logger=True,
                batch_size=batch_size,
            )
        elif split == "val":
            self.log(
                f"{split}/{source}/{name}",
                value,
                prog_bar=True,
                on_step=False,
                on_epoch=True,
                logger=True,
                batch_size=batch_size,
                sync_dist=True,
            )

    def log_dict_by_split(
        self,
        split: Split,
        source: str,
        dictionary: Mapping[str, _METRIC],
        batch_size: int | None = None,
    ) -> None:
        """Convenience method for logging multiple metrics simultaneously."""
        if split == "train":
            self.log_dict(
                {
                    f"{split}/{source}/{name}": metric
                    for name, metric in dictionary.items()
                },
                prog_bar=True,
                on_step=True,
                on_epoch=False,
                logger=True,
                batch_size=batch_size,
            )
        elif split == "val":
            self.log_dict(
                {
                    f"{split}/{source}/{name}": metric
                    for name, metric in dictionary.items()
                },
                prog_bar=True,
                on_step=False,
                on_epoch=True,
                logger=True,
                batch_size=batch_size,
                sync_dist=True,
            )

    def _process_and_log_losses(
        self,
        split: Split,
        losses: dict[str | None, dict[str, torch.Tensor]],
        batch_sizes: dict[str | None, int],
    ) -> torch.Tensor:
        """Logs single source losses and metrics and returns the mean of all loss components."""

        # simple single-source path
        # data batches from unnamed sources look like {None: single_source_batch}
        if None in losses:
            self.log_dict_by_split(
                split,
                "total",
                losses[None],
                batch_sizes[None],
            )
            return losses[None]["loss"]

        # for multiple datasets we log the total loss (optimisation target)
        # and, per source, each metric or loss component
        total_loss_list: list[torch.Tensor] = []

        for source, single_source_losses in losses.items():
            if source is None:
                continue

            single_source_batch_size = batch_sizes[source]

            self.log_dict_by_split(
                split,
                source,  # pyright: ignore
                single_source_losses,
                single_source_batch_size,
            )

            if "loss" in single_source_losses:
                total_loss_list.append(single_source_losses["loss"])

        total_loss = torch.stack(total_loss_list).mean()
        total_batch_size = sum(batch_sizes.values())

        self.log_by_split(
            split,
            "total",
            "loss",
            total_loss,
            batch_size=total_batch_size,
        )

        return total_loss



from collections.abc import Iterator, Mapping, Sequence
from os import PathLike as OSPathLike
from typing import (
    TypeVar,
    Union,
)

import fsspec

from datasets import Dataset, DatasetDict

# A list of lists of strings or Nones
DisplayNames = list[list[str | None]]

FileSystem = fsspec.AbstractFileSystem
ExamplesGenerator = Iterator[tuple[int | str, dict]]

# This includes str, pathlib.Path, cloudpathlib.CloudPath, ...
PathLike = Union[str, OSPathLike]

# Huggingface's DataFiles
HFDataFiles = Union[str, Sequence[str], Mapping[str, str | Sequence[str]]]
# An enhanced version accepting PathLike objects too
DataFiles = Union[
    PathLike,
    Sequence[PathLike],
    Mapping[str, PathLike | Sequence[PathLike]],
]

DatasetType = TypeVar("DatasetType", Dataset, DatasetDict)



"""
Tests for API features advertised as part of the Metric protocol.
"""

from typing import Protocol

from molflux.metrics.metric import Metric


def test_class_is_protocol():
    """That it is a protocol."""
    assert issubclass(type(Metric), type(Protocol))


def test_has_init():
    """That classes implementing the protocol can be initialised."""
    assert hasattr(Metric, "__init__")
    assert callable(Metric.__init__)


def test_has_len():
    """That it is sized."""
    assert hasattr(Metric, "__len__")
    assert callable(Metric.__len__)


def test_has_metadata():
    """That has a metadata property."""
    assert hasattr(Metric, "metadata")
    assert not callable(Metric.metadata)


def test_has_name():
    """That has a name property."""
    assert hasattr(Metric, "name")
    assert not callable(Metric.name)


def test_has_tag():
    """That has a tag property."""
    assert hasattr(Metric, "tag")
    assert not callable(Metric.tag)


def test_has_state():
    """That has a state property."""
    assert hasattr(Metric, "state")
    assert not callable(Metric.state)


def test_has_add_batch():
    """That has an add_batch method."""
    assert hasattr(Metric, "add_batch")
    assert callable(Metric.add_batch)


def test_has_compute():
    """That has a compute method."""
    assert hasattr(Metric, "compute")
    assert callable(Metric.compute)


def test_has_reset_state():
    """That has a reset_state method."""
    assert hasattr(Metric, "reset_state")
    assert callable(Metric.reset_state)


def test_has_udpate_state():
    """That has an update_state method."""
    assert hasattr(Metric, "update_state")
    assert callable(Metric.update_state)



from collections.abc import Iterator
from typing import Any

import numpy as np

from molflux.splits.bases import SplittingStrategyBase
from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.strategies.core._ordered_split_utils import (
    find_begin_end_indices,
    force_min_test_size,
    force_same_set_same_group,
    make_sorted_groups_indices,
    validate_ordered_strategy_arguments,
)
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable

_DESCRIPTION = """
A split based on the ordering of rows defined by the group column stratified by the target column. This can be used for any sort of temporal splits
which are stratified by membership to a set of targets (projects). Effectively, this is a sorting of the dataset by the groups column, stratifying by
the targets, and then doing a linear split.

Note: Will treat datapoints with no groups (None timestamps) as the earliest group.
"""


class StratifiedOrderedSplit(SplittingStrategyBase):
    def _info(self) -> SplittingStrategyInfo:
        return SplittingStrategyInfo(
            description=_DESCRIPTION,
        )

    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        *,
        train_fraction: float | list[float] = 0.8,
        gap_train_validation_fraction: float | list[float] | None = None,
        validation_fraction: float | list[float] = 0.1,
        gap_validation_test_fraction: float | list[float] | None = None,
        test_fraction: float | list[float] = 0.1,
        gap_test_end_fraction: float | list[float] | None = None,
        undefined_groups_in_train: bool = True,
        stratify_targets_to_ignore: list | None = None,
        min_test_size_per_target: int | list[int] | None = None,
        same_group_same_set: bool = True,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Args:
            dataset: The data to be split.
            y: The target variable for supervised learning problems (in case of a temporal split for
                project data, this could correspond to the project_id column). This splitting strategy tries to preserve
                the percentage of samples for each target.
            groups: Group labels for the samples used while splitting the dataset (in case of a temporal split for
                project data, this could correspond to the time_stamp column). This is used to order the samples
                before splitting.
            train_fraction (float or list): The proportion of the dataset to include in the train split.
                If a list is specified, split returns multiple folds with the specified fraction.
                Might not always be possible to get the exact fraction.
                Defaults to 0.8.
            gap_train_validation_fraction (Optional): Fraction of data to leave out between train and validation.
                Must be a float or a list, same as the train, validation fractions.
                Defaults to None
            validation_fraction: The proportion of the dataset to include in the validation split.
                If a list is specified, split returns multiple folds with the specified fraction.
                Might not always be possible to get the exact fraction.
                Defaults to 0.1.
            gap_validation_test_fraction (Optional): Fraction of data to leave out between validation and test.
                Must be a float or a list, same as the validation, test fractions.
                Defaults to None
            test_fraction: The proportion of the dataset to include in the test split.
                If a list is specified, split returns multiple folds with the specified fraction.
                Might not always be possible to get the exact fraction.
                Defaults to 0.1.
            gap_test_end_fraction (Optional): Fraction of data to leave out after test ste.
                Must be a float or a list, same as the test fractions.
                Defaults to None
            undefined_groups_in_train: If True, 'train_fraction' is computed on the dataset with None-groups removed,
                the resulting train-dataset might thus have a larger actual train_fraction than specified. Defaults to True.
            stratify_targets_to_ignore: The targets to ignore while doing stratification. Will be added to train. Defaults to None.
            min_test_size_per_target: The minimum number of datapoints in the test set of each target. Overrides the ratios.
                Can be an int for all the folds, or a list of ints with the same length as the folds. Defaults to None.
            same_group_same_set: Forces datapoints with the same group to be in the same set (will not respect specified
                ratios). If set to False, the specified ratios are respected, but which part of the group ends up in the
                 different sets will be based on the index order of the dataset (not determinisic if dataset is shuffled).
                Defaults to True.


        Examples:
            >>> from molflux.splits import load_splitting_strategy
            >>> strategy = load_splitting_strategy("stratified_ordered_split")
            >>>
            >>> x = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            >>> groups = [0, 0, 0, 2, 2, 2, 1, 1, 1, 1]
            >>> targets = ["a", "a", "b", "a", "b", "a", "a", "b", "a", "b"]
            >>>
            >>> train_idxs, val_idxs, test_idxs = next(strategy.split(x, groups=groups, y=targets))
            >>> sorted(train_idxs), sorted(val_idxs), sorted(test_idxs)
            ([0, 1, 2, 6, 7, 8, 9], [], [3, 4, 5])
        """

        # verify that all arguments are valid (and massage them into standard form
        validated_dict = validate_ordered_strategy_arguments(
            groups=groups,
            train_fraction=train_fraction,
            gap_train_validation_fraction=gap_train_validation_fraction,
            validation_fraction=validation_fraction,
            gap_validation_test_fraction=gap_validation_test_fraction,
            test_fraction=test_fraction,
            gap_test_end_fraction=gap_test_end_fraction,
            stratified_bool=True,
            y=y,
            stratify_targets_to_ignore=stratify_targets_to_ignore,
            min_test_size=min_test_size_per_target,
        )
        train_fraction_list = validated_dict["train_fraction_list"]
        gap_train_validation_fraction_list = validated_dict[
            "gap_train_validation_fraction_list"
        ]
        validation_fraction_list = validated_dict["validation_fraction_list"]
        gap_validation_test_fraction_list = validated_dict[
            "gap_validation_test_fraction_list"
        ]
        test_fraction_list = validated_dict["test_fraction_list"]
        gap_test_end_fraction_list = validated_dict["gap_test_end_fraction_list"]
        groups_array = validated_dict["groups_array"]
        index_array = validated_dict["index_array"]
        targets_array = validated_dict["targets_array"]
        set_of_targets = validated_dict["set_of_targets"]
        min_test_size_per_target_list = validated_dict["min_test_size_list"]

        # the indices of targets to ignore
        targets_to_ignore_indices = index_array[
            np.array([targets_array != x for x in set_of_targets]).all(0)
        ]

        # iterate over number of folds
        for ii in range(len(train_fraction_list)):
            # fractions for this split
            split_train_fraction = train_fraction_list[ii]
            split_gap_train_val_fraction = gap_train_validation_fraction_list[ii]
            split_validation_fraction = validation_fraction_list[ii]
            split_gap_val_test_fraction = gap_validation_test_fraction_list[ii]
            split_test_fraction = test_fraction_list[ii]
            split_gap_test_end_fraction = gap_test_end_fraction_list[ii]
            split_min_test_size_per_target = min_test_size_per_target_list[ii]

            # empty indices
            train_indices = []
            validation_indices = []
            test_indices = []

            # add everything not in targets to train
            train_indices += targets_to_ignore_indices.tolist()

            for target in set_of_targets:
                sorted_groups_indices_dict = make_sorted_groups_indices(
                    groups_array=groups_array,
                    index_array=index_array,
                    targets_array=targets_array,
                    target=target,
                )
                sorted_groups_with_nones = sorted_groups_indices_dict[
                    "sorted_groups_with_nones"
                ]
                sorted_indices_without_nones = sorted_groups_indices_dict[
                    "sorted_indices_without_nones"
                ]
                sorted_indices_with_nones = sorted_groups_indices_dict[
                    "sorted_indices_with_nones"
                ]

                begin_end_dict = find_begin_end_indices(
                    sorted_indices_without_nones=sorted_indices_without_nones,
                    sorted_indices_with_nones=sorted_indices_with_nones,
                    split_train_fraction=split_train_fraction,
                    split_gap_train_val_fraction=split_gap_train_val_fraction,
                    split_validation_fraction=split_validation_fraction,
                    split_gap_val_test_fraction=split_gap_val_test_fraction,
                    split_test_fraction=split_test_fraction,
                    split_gap_test_end_fraction=split_gap_test_end_fraction,
                    undefined_groups_in_train=undefined_groups_in_train,
                )
                train_end = begin_end_dict["train_end"]
                validation_begin = begin_end_dict["validation_begin"]
                validation_end = begin_end_dict["validation_end"]
                test_begin = begin_end_dict["test_begin"]
                test_end = begin_end_dict["test_end"]

                # set required min test size (is specified and test_size is not enough)
                if split_min_test_size_per_target is not None:
                    begin_end_dict = force_min_test_size(
                        min_test_size=split_min_test_size_per_target,
                        train_end=train_end,
                        validation_begin=validation_begin,
                        validation_end=validation_end,
                        test_begin=test_begin,
                        test_end=test_end,
                    )
                    train_end = begin_end_dict["train_end"]
                    validation_begin = begin_end_dict["validation_begin"]
                    validation_end = begin_end_dict["validation_end"]
                    test_begin = begin_end_dict["test_begin"]
                    test_end = begin_end_dict["test_end"]

                # if same group in same set
                if same_group_same_set:
                    # change sets to the nearest whole group
                    train_end = force_same_set_same_group(
                        sorted_groups_with_nones,
                        train_end,
                    )
                    validation_begin = force_same_set_same_group(
                        sorted_groups_with_nones,
                        validation_begin,
                    )
                    validation_end = force_same_set_same_group(
                        sorted_groups_with_nones,
                        validation_end,
                    )
                    test_begin = force_same_set_same_group(
                        sorted_groups_with_nones,
                        test_begin,
                    )
                    test_end = force_same_set_same_group(
                        sorted_groups_with_nones,
                        test_end,
                    )

                target_train_indices = sorted_indices_with_nones[:train_end].tolist()
                target_validation_indices = sorted_indices_with_nones[
                    validation_begin:validation_end
                ].tolist()
                target_test_indices = sorted_indices_with_nones[
                    test_begin:test_end
                ].tolist()

                # add to train, val, test indices
                train_indices += target_train_indices
                validation_indices += target_validation_indices
                test_indices += target_test_indices

            train_indices = list(
                set(train_indices).difference(*(validation_indices, test_indices)),
            )
            validation_indices = list(set(validation_indices).difference(test_indices))
            test_indices = list(set(test_indices))

            yield train_indices, validation_indices, test_indices
